from typing import (
    Iterable,
    Optional,
)
from uuid import UUID

from requests import Request

from encord.http.v2.api_client import ApiClient
from encord.objects.ml_models import (
    ModelCreateRequest,
    ModelIterationTrainingData,
    ModelIterationTrainingDataListRequest,
    ModelsListOrderBy,
    ModelsListRequest,
    ModelTrainingRequest,
    ModelTrainingResponse,
    ModelUpdateRequest,
    ModelWithIterations,
    PredictionClassificationRequest,
    PredictionClassificationResponse,
    PredictionInstanceSegmentationRequest,
    PredictionInstanceSegmentationResponse,
    PredictionObjectDetectionRequest,
    PredictionObjectDetectionResponse,
    ProjectModelAttachRequest,
    ProjectModelsListRequest,
    ProjectModelUpdateRequest,
    ProjectModelWithIterations,
)


def _api_ml_create_model(
    *,
    client: ApiClient,
    body: ModelCreateRequest,
) -> ModelWithIterations:
    return client._request(
        Request(
            method="post",
            url=f"{client._domain}/v2/public/ml-models",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=ModelWithIterations,
    )


def _api_ml_get_model(
    model_uuid: UUID,
    *,
    client: ApiClient,
) -> ModelWithIterations:
    return client._request(
        Request(
            method="get",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}",
        ).prepare(),
        result_type=ModelWithIterations,
    )


def _api_ml_list_models(
    *,
    client: ApiClient,
    order_by: ModelsListOrderBy,
    order_asc: bool,
    query: Optional[str],
) -> Iterable[ModelWithIterations]:
    yield from client.get_paged_iterator(
        "ml-models",
        params=ModelsListRequest(
            order_by=order_by,
            order_asc=order_asc,
            query=query,
            page_token=None,
        ),
        result_type=ModelWithIterations,
    )


def _api_ml_delete_model(
    model_uuid: UUID,
    *,
    client: ApiClient,
) -> None:
    return client._request(
        Request(
            method="delete",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}",
        ).prepare(),
        result_type=None,
    )


def _api_ml_update_model(
    model_uuid: UUID,
    *,
    client: ApiClient,
    body: ModelUpdateRequest,
) -> ModelWithIterations:
    return client._request(
        Request(
            method="patch",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=ModelWithIterations,
    )


def _api_ml_create_training_job(
    model_uuid: UUID,
    *,
    client: ApiClient,
    body: ModelTrainingRequest,
) -> UUID:
    return client._request(
        Request(
            method="post",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}/trainings",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=UUID,
    )


def _api_ml_get_training_status(
    model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
    timeout_seconds: Optional[int] = 0,
) -> ModelTrainingResponse:
    return client._request(
        Request(
            method="get",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}/iterations/{training_uuid}/training",
            params={
                k: v
                for k, v in {
                    "timeoutSeconds": timeout_seconds,
                }.items()
                if v is not None
            },
        ).prepare(),
        result_type=ModelTrainingResponse,
    )


def _api_ml_get_training_data(
    model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
) -> Iterable[ModelIterationTrainingData]:
    yield from client.get_paged_iterator(
        f"ml-models/{model_uuid}/iterations/{training_uuid}/training-data",
        params=ModelIterationTrainingDataListRequest(
            page_token=None,
        ),
        result_type=ModelIterationTrainingData,
    )


def _api_ml_get_weights_download_link(
    model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
) -> str:
    return client._request(
        Request(
            method="get",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}/iterations/{training_uuid}/weights-link",
        ).prepare(),
        result_type=str,
    )


def _api_ml_delete_training_iteration(
    model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
) -> None:
    return client._request(
        Request(
            method="delete",
            url=f"{client._domain}/v2/public/ml-models/{model_uuid}/iterations/{training_uuid}",
        ).prepare(),
        result_type=None,
    )


def _api_project_create_model_attachment(
    project_uuid: UUID,
    *,
    client: ApiClient,
    body: ProjectModelAttachRequest,
) -> ProjectModelWithIterations:
    return client._request(
        Request(
            method="post",
            url=f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=ProjectModelWithIterations,
    )


def _api_project_list_model_attachments(
    project_uuid: UUID,
    *,
    client: ApiClient,
) -> Iterable[ProjectModelWithIterations]:
    yield from client.get_paged_iterator(
        f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments",
        params=ProjectModelsListRequest(
            page_token=None,
        ),
        result_type=ProjectModelWithIterations,
    )


def _api_project_update_model_attachment(
    project_uuid: UUID,
    project_model_uuid: UUID,
    *,
    client: ApiClient,
    body: ProjectModelUpdateRequest,
) -> ProjectModelWithIterations:
    return client._request(
        Request(
            method="patch",
            url=f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments/{project_model_uuid}",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=ProjectModelWithIterations,
    )


def _api_project_delete_model_attachment(
    project_uuid: UUID,
    project_model_uuid: UUID,
    *,
    client: ApiClient,
) -> None:
    return client._request(
        Request(
            method="delete",
            url=f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments/{project_model_uuid}",
        ).prepare(),
        result_type=None,
    )


def _api_project_predict_classification(
    project_uuid: UUID,
    project_model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
    body: PredictionClassificationRequest,
) -> PredictionClassificationResponse:
    return client._request(
        Request(
            method="post",
            url=f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments/{project_model_uuid}/iterations/{training_uuid}/classification/predict",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=PredictionClassificationResponse,
    )


def _api_project_predict_instance_segmentation(
    project_uuid: UUID,
    project_model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
    body: PredictionInstanceSegmentationRequest,
) -> PredictionInstanceSegmentationResponse:
    return client._request(
        Request(
            method="post",
            url=f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments/{project_model_uuid}/iterations/{training_uuid}/instance-segmentation/predict",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=PredictionInstanceSegmentationResponse,
    )


def _api_project_predict_object_detection(
    project_uuid: UUID,
    project_model_uuid: UUID,
    training_uuid: UUID,
    *,
    client: ApiClient,
    body: PredictionObjectDetectionRequest,
) -> PredictionObjectDetectionResponse:
    return client._request(
        Request(
            method="post",
            url=f"{client._domain}/v2/public/projects/{project_uuid}/models-attachments/{project_model_uuid}/iterations/{training_uuid}/object-detection/predict",
            json=client._serialise_payload(body),
        ).prepare(),
        result_type=PredictionObjectDetectionResponse,
    )



import uuid
from datetime import datetime
from typing import Dict, List, Optional
from uuid import UUID

from encord.orm.base_dto import BaseDTO, Field, dto_validator


class GetPresetParams(BaseDTO):
    top_level_folder_uuid: Optional[UUID] = Field(default=None, alias="topLevelFolderUuid")
    preset_uuids: Optional[List[UUID]] = Field(default=[], alias="uuids")
    page_token: Optional[str] = Field(default=None, alias="pageToken")
    page_size: Optional[int] = Field(default=None, alias="pageSize")


class FilterPreset(BaseDTO):
    uuid: uuid.UUID
    name: str
    description: Optional[str]
    created_at: Optional[datetime] = Field(default=None, alias="createdAt")
    last_updated_at: Optional[datetime] = Field(default=None, alias="lastUpdatedAt")


class GetProjectFilterPresetParams(BaseDTO):
    preset_uuids: Optional[List[uuid.UUID]] = Field(default=[])
    page_token: Optional[str] = Field(default=None)
    page_size: Optional[int] = Field(default=None)


class ProjectFilterPreset(BaseDTO):
    preset_uuid: uuid.UUID = Field(alias="presetUuid")
    name: str
    created_at: Optional[datetime] = Field(default=None)
    updated_at: Optional[datetime] = Field(default=None)


class FilterDefinition(BaseDTO):
    filters: List[Dict] = Field(default_factory=list)


# Note alias is strictly required as these are stored in the Annotate DB as unstructured objects
# Stored not in camelCase like most Models
class IndexFilterPresetDefinition(BaseDTO):
    local_filters: Dict[str, FilterDefinition] = Field(
        default_factory=lambda: {str(uuid.UUID(int=0)): FilterDefinition()},
        alias="local_filters",
    )
    global_filters: FilterDefinition = Field(default_factory=FilterDefinition, alias="global_filters")


class ActiveFilterPresetDefinition(BaseDTO):
    local_filters: Dict[str, FilterDefinition] = Field(
        default_factory=lambda: {str(uuid.UUID(int=0)): FilterDefinition()},
    )
    global_filters: FilterDefinition = Field(default_factory=FilterDefinition)


class GetPresetsResponse(BaseDTO):
    results: List[FilterPreset]


class CreatePresetParams(BaseDTO):
    top_level_folder_uuid: UUID = Field(default=UUID(int=0), alias="topLevelFolderUuid")


class IndexCreatePresetPayload(BaseDTO):
    name: str
    filter_preset_json: Dict
    description: Optional[str] = ""


class ActiveCreatePresetPayload(BaseDTO):
    name: str
    filter_preset_json: Dict


class IndexUpdatePresetPayload(BaseDTO):
    name: Optional[str] = None
    description: Optional[str] = ""
    filter_preset: Optional[IndexFilterPresetDefinition] = None


class ActiveUpdatePresetPayload(BaseDTO):
    name: Optional[str] = None
    filter_preset: Optional[ActiveFilterPresetDefinition] = None



from encord.objects.attributes import ChecklistAttribute, RadioAttribute, TextAttribute
from encord.objects.classification import Classification
from encord.objects.classification_instance import ClassificationInstance
from encord.objects.common import Shape
from encord.objects.metadata import DICOMSeriesMetadata, DICOMSliceMetadata
from encord.objects.ontology_labels_impl import LabelRowV2
from encord.objects.ontology_object import Object
from encord.objects.ontology_object_instance import AnswerForFrames, ObjectInstance
from encord.objects.ontology_structure import OntologyStructure
from encord.objects.options import FlatOption, NestableOption, Option



import inspect
import json
import uuid
from http import HTTPStatus
from typing import Callable, Dict, Iterator, List, Optional, Sequence, Type, TypeVar, Union
from urllib.parse import urljoin

import requests
from pydantic import BaseModel
from requests import PreparedRequest, Response

from encord.configs import Config
from encord.exceptions import EncordException, RequestException
from encord.http.common import (
    HEADER_CLOUD_TRACE_CONTEXT,
    RequestContext,
)
from encord.http.utils import create_new_session
from encord.http.v2.error_utils import handle_error_response
from encord.http.v2.payloads import Page
from encord.orm.base_dto import BaseDTO, BaseDTOInterface

T = TypeVar("T", bound=Union[Sequence[BaseDTOInterface], BaseDTOInterface, uuid.UUID, int, str])


class ApiClient:
    def __init__(self, config: Config):
        self._config = config
        self._domain = self._config.domain
        self._base_path = "v2/public/"
        self._bound_callbacks: Dict[Callable, Callable] = {}

    @staticmethod
    def _exception_context(request: requests.PreparedRequest) -> RequestContext:
        try:
            x_cloud_trace_context = request.headers.get(HEADER_CLOUD_TRACE_CONTEXT)
            if x_cloud_trace_context is None:
                return RequestContext()

            x_cloud_trace_context = x_cloud_trace_context.split(";")[0]
            trace_id, span_id = (x_cloud_trace_context.split("/") + [None, None])[:2]
            return RequestContext(trace_id=trace_id, span_id=span_id)
        except Exception:
            return RequestContext()

    def _build_url(self, path: str) -> str:
        if path.startswith("/"):
            path = path[1:]
        url = urljoin(self._domain, urljoin(self._base_path, path))
        if url.endswith("/"):
            url = url[:-1]
        return url

    def get_bound_operation(self, operation: Callable) -> Callable:
        """Wrap a function to bind it to the current API client instance (via a named parameter). This is useful for
        bundling, as the 'Bundle' groups operations based on the 'operation' function, which means that if
        you use an entity object method as the operation, then it will not be grouped with other similar operations.
        """
        if "api_client" in inspect.signature(operation).parameters:
            if wrapped := self._bound_callbacks.get(operation):
                return wrapped

            def wrapped_callback(*args, **kwargs):
                return operation(*args, api_client=self, **kwargs)

            self._bound_callbacks[operation] = wrapped_callback

            return wrapped_callback
        else:
            raise RuntimeError(f"Operation {operation} does not have an 'api_client' parameter")

    def get(self, path: str, params: Optional[BaseDTO], result_type: Type[T], allow_none: bool = False) -> T:
        return self._request_without_payload("GET", path, params, result_type, allow_none=allow_none)

    def get_paged_iterator(
        self,
        path: str,
        params: BaseDTO,
        result_type: Type[T],
        allow_none: bool = False,
    ) -> Iterator[T]:
        if not hasattr(params, "page_token"):
            raise ValueError("params must have a page_token attribute for paging to work")

        while True:
            #  Pydantic is magic and relies on under-specified parts of the type system
            #  MyPy doesn't like this (insists on 'type erasure'), but it works because
            #  in reality the type is not erased and the generic parameters are available
            page = self.get(
                path,
                params=params,
                result_type=Page[result_type],  # type: ignore[valid-type]
                allow_none=allow_none,
            )

            yield from page.results

            if page.next_page_token is not None:
                params.page_token = page.next_page_token
            else:
                break

    def delete(
        self, path: str, params: Optional[BaseDTO], result_type: Optional[Type[T]] = None, allow_none: bool = False
    ) -> T:
        return self._request_without_payload("DELETE", path, params, result_type, allow_none=allow_none)

    def post(
        self,
        path: str,
        params: Optional[BaseDTO],
        payload: Union[BaseDTO, Sequence[BaseDTO], None],
        result_type: Optional[Type[T]],
    ) -> T:
        return self._request_with_payload("POST", path, params, payload, result_type)

    def put(
        self,
        path: str,
        params: Optional[BaseDTO],
        payload: Union[BaseDTO, Sequence[BaseDTO], None],
    ) -> None:
        self._request_with_payload("PUT", path, params, payload, None, allow_none=True)

    def patch(
        self, path: str, params: Optional[BaseDTO], payload: Optional[BaseDTO], result_type: Optional[Type[T]]
    ) -> T:
        return self._request_with_payload("PATCH", path, params, payload, result_type)

    def _serialise_payload(self, payload: Union[BaseDTO, Sequence[BaseDTO], None]) -> Union[List[Dict], Dict, None]:
        if isinstance(payload, list):
            return [p.to_dict() for p in payload]
        elif isinstance(payload, BaseDTO):
            return payload.to_dict()
        elif isinstance(payload, BaseModel):
            # use new pydantic v2 function if it exists, otherwise use fallback
            if hasattr(payload, "model_dump"):
                return payload.model_dump(mode="json")
            else:
                return json.loads(payload.json())
        elif payload is None:
            return None
        else:
            raise ValueError(f"Unsupported payload type: {type(payload)}")

    def _request_with_payload(
        self,
        method: str,
        path: str,
        params: Optional[BaseDTO],
        payload: Union[BaseDTO, Sequence[BaseDTO], None],
        result_type: Optional[Type[T]],
        allow_none: bool = False,
    ) -> T:
        params_dict = params.to_dict() if params is not None else None
        payload_serialised = self._serialise_payload(payload)

        req = requests.Request(
            method=method,
            url=self._build_url(path),
            params=params_dict,
            json=payload_serialised,
        ).prepare()

        return self._request(req, result_type=result_type, allow_none=allow_none)  # type: ignore

    def _request_without_payload(
        self,
        method: str,
        path: str,
        params: Optional[BaseDTO],
        result_type: Optional[Type[T]],
        allow_none: bool = False,
    ) -> T:
        params_dict = params.to_dict() if params is not None else None

        req = requests.Request(method=method, url=self._build_url(path), params=params_dict).prepare()

        return self._request(req, result_type=result_type, allow_none=allow_none)  # type: ignore

    def _request(self, req: PreparedRequest, result_type: Optional[Type[T]], allow_none: bool = False):
        req = self._config.define_headers_v2(req)

        timeouts = (self._config.connect_timeout, self._config.read_timeout)
        req_settings = self._config.requests_settings
        with create_new_session(
            max_retries=req_settings.max_retries,
            backoff_factor=req_settings.backoff_factor,
            connect_retries=req_settings.connection_retries,
        ) as session:
            context = self._exception_context(req)

            try:
                res = session.send(req, timeout=timeouts)
            except Exception as e:
                raise RequestException(f"Request session.send failed {req.method=} {req.url=}", context=context) from e

            if res.status_code not in [
                HTTPStatus.OK,
                HTTPStatus.NO_CONTENT,  # 204 status code will raise error for sdk versions <= 0.1.147
            ]:
                self._handle_error(res, context)

            if res.status_code == HTTPStatus.NO_CONTENT:
                return None

            try:
                res_json = res.json()
            except Exception as e:
                raise RequestException(f"Error parsing JSON response: {res.text}", context=context) from e

            if result_type is None or (res_json is None and allow_none):
                return None
            if result_type is int:
                return int(res_json)
            elif result_type is str:
                return str(res_json)
            elif result_type is uuid.UUID:
                return uuid.UUID(res_json)
            elif issubclass(result_type, BaseDTOInterface):
                return result_type.from_dict(res_json)
            elif issubclass(result_type, BaseModel):
                # use new pydantic v2 function if it exists, otherwise use fallback
                if hasattr(result_type, "model_validate"):
                    return result_type.model_validate(res_json)
                else:
                    return result_type.validate(res_json)
            else:
                raise ValueError(f"Unsupported result type {result_type}")

    @staticmethod
    def _handle_error(response: Response, context: RequestContext):
        try:
            description = response.json()
            handle_error_response(response.status_code, context=context, message=description.get("message"))
        except EncordException as e:
            raise e
        except Exception:
            handle_error_response(response.status_code, context=context)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from enum import auto
from typing import Dict, List, Optional, Union
from uuid import UUID

from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO, Field
from encord.orm.dataset import DataUnitError, LongPollingStatus

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore[assignment]


class StorageItemType(CamelStrEnum):
    VIDEO = auto()
    IMAGE = auto()
    IMAGE_GROUP = auto()
    IMAGE_SEQUENCE = auto()
    DICOM_FILE = auto()
    DICOM_SERIES = auto()
    AUDIO = auto()
    NIFTI = auto()
    PLAIN_TEXT = auto()
    PDF = auto()


class StorageUserRole(CamelStrEnum):
    USER = auto()
    ADMIN = auto()


class StorageLocationName(CamelStrEnum):
    ENCORD = auto()
    GCP = auto()
    S3 = auto()
    AZURE = auto()
    OPEN_TELEKOM = auto()  # TODO: Remove when enough people have updated their sdk to include S3_COMPATIBLE below
    DIRECT_ACCESS = auto()
    S3_COMPATIBLE = auto()


class PathElement(BaseDTO):
    uuid: UUID
    parent_uuid: Optional[UUID]
    name: str
    synced_dataset_hash: Optional[UUID]


class StorageFolder(BaseDTO):
    uuid: UUID
    parent: Optional[UUID]
    name: str
    description: str
    client_metadata: Optional[str]
    owner: str
    created_at: datetime
    last_edited_at: datetime
    user_role: StorageUserRole
    synced_dataset_hash: Optional[UUID]
    path_to_root: List[PathElement]


class StorageItem(BaseDTO):
    uuid: UUID
    parent: UUID
    item_type: StorageItemType
    name: str
    description: str
    client_metadata: str
    owner: str
    created_at: datetime
    last_edited_at: datetime
    is_tombstone: bool = False
    """This item has been deleted but the link is retained for consistency reasons.
    Mostly for items in the 'cloud linked folders' that are referenced but aren't present after a re-sync"""
    is_placeholder: bool = False
    """This item has been added to the folder but isn't fully processed yet"""
    backed_data_units_count: int
    storage_location: StorageLocationName
    integration_hash: Optional[UUID]
    url: Optional[str]
    signed_url: Optional[str]
    file_size: Optional[int]
    mime_type: Optional[str]
    duration: Optional[float]
    fps: Optional[float]
    height: Optional[int]
    width: Optional[int]
    dicom_instance_uid: Optional[str]
    dicom_study_uid: Optional[str]
    dicom_series_uid: Optional[str]
    frame_count: Optional[int]
    audio_sample_rate: Optional[int]
    audio_bit_depth: Optional[int]
    audio_codec: Optional[str]
    audio_num_channels: Optional[int]


class StorageItemInaccessible(BaseDTO):
    uuid: UUID


class CreateStorageFolderPayload(BaseDTO):
    name: str
    description: Optional[str]
    parent: Optional[UUID]
    client_metadata: Optional[str]


class UploadSignedUrlsPayload(BaseDTO):
    item_type: StorageItemType
    count: int
    frames_subfolder_name: Optional[str]


class UploadSignedUrl(BaseDTO):
    item_uuid: UUID
    object_key: str
    signed_url: str


class StorageItemWithName(BaseDTO):
    item_uuid: UUID
    name: str


class UploadLongPollingState(BaseDTO):
    """Response of the upload job's long polling request.

    Note: An upload job consists of job units, where job unit could be
    either a video, image group, dicom series, or a single image.
    """

    status: LongPollingStatus
    """Status of the upload job. Documented in detail in :meth:`encord.orm.dataset.LongPollingStatus`"""

    items_with_names: List[StorageItemWithName]
    """Information about data which was added to the folder."""

    errors: List[str]
    """Stringified list of exceptions."""

    units_pending_count: int
    """Number of upload job units that have pending status."""

    units_done_count: int
    """Number of upload job units that have done status."""

    units_error_count: int
    """Number of upload job units that have error status."""

    units_cancelled_count: int
    """Number of upload job units that have been cancelled."""

    unit_errors: List[DataUnitError]
    """Structured list of per-item upload errors. See :class:`DataUnitError` for more details."""

    file_name: Optional[str] = None
    """Name of the JSON or CSV file that contained the list of URLs to ingest form the cloud bucket. Optional."""


class CustomerProvidedImageMetadata(BaseDTO):
    """Media metadata for an image file; if provided, Encord service will use the values here instead of scanning the files"""

    mime_type: str
    """MIME type of the image file (e.g. `image/jpeg` or `image/png`)."""
    file_size: int
    """Size of the image file in bytes."""
    height: int
    """Height of the image in pixels."""
    width: int
    """Width of the image in pixels."""


class CustomerProvidedVideoMetadata(BaseDTO):
    """Media metadata for a video file; if provided, Encord service will skip frame synchronisation checks
    and will use the values specified here to render the video in the label editor.
    """

    fps: float
    """Frame rate of the video in frames per second."""
    duration: float
    """Video duration in (float) seconds."""
    width: int
    """Width of the video in pixels."""
    height: int
    """Height of the video in pixels."""
    file_size: int
    """Size of the video file in bytes."""
    mime_type: str
    """MIME type of the video file (e.g. `video/mp4` or `video/webm`)."""


class CustomerProvidedAudioMetadata(BaseDTO):
    """Media metadata for an audio file; if provided, Encord service will use the values here instead of scanning the files"""

    duration: float
    """Audio duration in (float) seconds."""
    file_size: int
    """Size of the audio file in bytes."""
    mime_type: str
    """MIME type of the audio file (for example: `audio/mpeg` or `audio/wav`)."""
    sample_rate: int
    """Sample rate (int) in Hz."""
    bit_depth: int
    """Size of each sample (int) in bits."""
    codec: str
    """Codec (for example: mp3, pcm)."""
    num_channels: int
    """Number of channels"""


class CustomerProvidedDicomSeriesDicomFileMetadata(BaseDTO):
    """Metadata for a DICOM file containing required DICOM tags and their values.
    This metadata is used to validate and process DICOM files without needing to access the actual files.

    The `tags` dictionary must contain all required DICOM tags as keys, though their corresponding values may be None.
    Tags should be provided in the format returned by pydicom to_json_dict() method.

    Required tags (using standard DICOM tag numbers):
        - 00080018: SOPInstanceUID
        - 00100020: PatientID
        - 00180050: SliceThickness
        - 00181114: EstimatedRadiographicMagnificationFactor
        - 00181164: ImagerPixelSpacing
        - 0020000D: StudyInstanceUID
        - 0020000E: SeriesInstanceUID
        - 00200013: InstanceNumber
        - 00200032: ImagePositionPatient
        - 00200037: ImageOrientationPatient
        - 00209113: PlanePositionSequence
        - 00209116: PlaneOrientationSequence
        - 00280004: PhotometricInterpretation
        - 00280008: NumberOfFrames
        - 00280010: Rows
        - 00280011: Columns
        - 00280030: PixelSpacing
        - 00281050: WindowCenter
        - 00281051: WindowWidth
        - 00289110: PixelMeasuresSequence
        - 52009229: SharedFunctionalGroupsSequence
        - 52009230: PerFrameFunctionalGroupsSequence

    Missing any of these tags as `tags` dictionary key will raise a validation error.
    """

    tags: Dict[str, Optional[Dict]]


class DataUploadImage(BaseDTO):
    """Data about a single image item to be registered with Encord service."""

    object_url: str
    """"URL of the image file to be registered."""
    title: Optional[str] = None
    """Title of the image item (derived from the URL if omitted)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the image item."""
    external_file_type: Literal["IMAGE"] = "IMAGE"
    """Type of the external file."""
    image_metadata: Optional[CustomerProvidedImageMetadata] = None
    """Optional media metadata of the image file (if provided). See :class:`CustomerProvidedImageMetadata` for more details."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadVideo(BaseDTO):
    """Data about a video item to be registered with Encord service."""

    object_url: str
    """URL of the video file to be registered."""
    title: Optional[str] = None
    """Title of the video item (derived from the URL if omitted)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the video item."""

    external_file_type: Literal["VIDEO"] = "VIDEO"
    """Type of the external file."""
    video_metadata: Optional[CustomerProvidedVideoMetadata] = None
    """Optional media metadata of the video file (if provided). See :class:`CustomerProvidedVideoMetadata` for more details."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadNifti(BaseDTO):
    """Data about a NIFTI item to be registered with Encord service."""

    object_url: str
    """URL of the NIFTI file to be registered."""
    title: Optional[str] = None
    """Title of the NIFTI item (derived from the URL if omitted)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the NIFTI item."""

    external_file_type: Literal["NIFTI"] = "NIFTI"
    """Type of the external file."""
    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadImageGroupImage(BaseDTO):
    """Data about a single image item to be used as a frame in an image group or sequence."""

    url: str
    """URL of the image file to be used as the frame in the image group."""
    title: Optional[str] = None
    """Title of the image item (derived from the URL if omitted)."""
    image_metadata: Optional[CustomerProvidedImageMetadata] = None
    """Optional media metadata of the image file (if provided). See :class:`CustomerProvidedImageMetadata` for more details."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadImageGroup(BaseDTO):
    """Data about an image group or image sequence item to be registered with Encord service."""

    images: List[DataUploadImageGroupImage]
    """List of images to be used as frames in the image group. See :class:`DataUploadImageGroupImage` for more details."""
    title: Optional[str]
    """Title of the image group item (requred if using cloud integration)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the image group item."""
    create_video: bool = False
    """If set to `True`, create an image sequence backed by a video file uploaded to the cloud storage."""

    external_file_type: Literal["IMG_GROUP"] = "IMG_GROUP"
    """Type of the external file."""
    cluster_by_resolution: bool = False
    """For system use only."""


class DataUploadImageGroupFromItems(BaseDTO):
    """Data about an image group item to be created from previously uploaded images."""

    image_items: List[UUID]
    """List of image items to be used as frames in the image group."""
    title: Optional[str]
    """Title of the image group item (required if using cloud integration)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the image group."""
    create_video: bool
    """If set to `True`, create an image sequence backed by a video file uploaded to the cloud storage."""
    video_url_prefix: Optional[str] = None
    """URL prefix for the video file to be created from the image group."""

    external_file_type: Literal["IMG_GROUP_FROM_ITEMS"] = "IMG_GROUP_FROM_ITEMS"
    """Type of the external file."""
    cluster_by_resolution: bool = False
    """For system use only."""


class DataUploadDicomSeriesDicomFile(BaseDTO):
    """Data about a single DICOM file to be used in a series item."""

    url: str
    """URL of the DICOM file to be registered with Encord service."""
    title: Optional[str]
    """Title of the DICOM file (derived from the URL if omitted)."""
    dicom_metadata: Optional[CustomerProvidedDicomSeriesDicomFileMetadata] = None
    """Optional media metadata of the DICOM file (if provided). See :class:`CustomerProvidedDicomSeriesDicomFileMetadata` for more details."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadDicomSeries(BaseDTO):
    """Data about a DICOM series item to be registered with Encord service."""

    dicom_files: List[DataUploadDicomSeriesDicomFile]
    """List of DICOM files to be used in the series item. See :class:`DataUploadDicomSeriesDicomFile` for more details."""
    title: Optional[str]
    """Title of the DICOM series item (required if using cloud integration)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the DICOM series item."""

    external_file_type: Literal["DICOM"] = "DICOM"
    """Type of the external file."""


class DataUploadText(BaseDTO):
    object_url: str
    """URL of the text (TXT, HTML, etc) file to be registered with Encord service."""
    title: Optional[str] = None
    """Title of the file (derived from the URL if omitted)."""
    client_metadata: dict = Field(default_factory=dict)
    """Custom metadata to be associated with the file."""

    external_file_type: Literal["PLAIN_TEXT"] = "PLAIN_TEXT"
    """Type of the external file."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadPDF(BaseDTO):
    object_url: str
    """URL of the PDF file to be registered with Encord service."""
    title: Optional[str] = None
    """Title of the file (derived from the URL if omitted)."""
    client_metadata: dict = Field(default_factory=dict)
    """Custom metadata to be associated with the file."""

    external_file_type: Literal["PDF"] = "PDF"
    """Type of the external file."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadAudio(BaseDTO):
    """Data about an audio item to be registered with Encord service."""

    object_url: str
    """URL of the audio file to be registered."""
    title: Optional[str] = None
    """Title of the audio item (derived from the URL if omitted)."""
    client_metadata: Dict = Field(default_factory=dict)
    """Custom metadata to be associated with the audio item."""

    audio_metadata: Optional[CustomerProvidedAudioMetadata] = None
    """Optional media metadata of the audio file (if provided). See :class:`CustomerProvidedAudioMetadata` for more details."""
    external_file_type: Literal["AUDIO"] = "AUDIO"
    """Type of the external file."""

    placeholder_item_uuid: Optional[UUID] = None
    """For system use only."""


class DataUploadItems(BaseDTO):
    """A collection of items to be registered with Encord service.

    A more structured alternative to using a JSON file.
    """

    videos: List[DataUploadVideo] = Field(default_factory=list)
    """List of video items to be registered. See :class:`DataUploadVideo` for more details."""

    image_groups: List[DataUploadImageGroup] = Field(default_factory=list)
    """List of image group items to be registered. See :class:`DataUploadImageGroup` for more details."""

    dicom_series: List[DataUploadDicomSeries] = Field(default_factory=list)
    """List of DICOM series items to be registered. See :class:`DataUploadDicomSeries` for more details."""

    images: List[DataUploadImage] = Field(default_factory=list)
    """List of image items to be registered. See :class:`DataUploadImage` for more details."""

    image_groups_from_items: List[DataUploadImageGroupFromItems] = Field(default_factory=list)
    """List of image group items to be created from previously uploaded images.
    See :class:`DataUploadImageGroupFromItems` for more details."""

    audio: List[DataUploadAudio] = Field(default_factory=list)
    """List of audio items to be registered. See :class:`DataUploadAudio` for more details."""

    nifti: List[DataUploadNifti] = Field(default_factory=list)
    """List of NIFTI items to be registered. See :class:`DataUploadNifti` for more details."""

    text: List[DataUploadText] = Field(default_factory=list)
    """List of text items to be registered. See :class:`DataUploadText` for more details."""

    pdf: List[DataUploadPDF] = Field(default_factory=list)
    """List of PDF items to be registered. See :class:`DataUploadPDF` for more details."""

    skip_duplicate_urls: bool = False
    """If set to `True`, Encord service will skip items with URLs that already exist in the same folder.
    Otherwise, duplicate items will be created."""

    upsert_metadata: bool = False
    """If set to `True`, Encord service will update metadata of existing items with the same URL in the same folder.
    This flag has no effect if `skip_duplicate_urls` is set to `False`."""


class DatasetDataLongPollingParams(BaseDTO):
    data_items: Optional[DataUploadItems]
    files: Optional[dict]
    integration_id: Optional[UUID]
    ignore_errors: bool
    folder_uuid: Optional[UUID]
    file_name: Optional[str]


class PostUploadJobParams(BaseDTO):
    data_items: Optional[DataUploadItems] = None
    external_files: Optional[dict] = None
    integration_hash: Optional[UUID] = None
    ignore_errors: bool = False
    file_name: Optional[str] = None


class GetUploadJobParams(BaseDTO):
    timeout_seconds: int = 60


class FoldersSortBy(CamelStrEnum):
    NAME = auto()
    CREATED_AT = auto()


class ListItemsParams(BaseDTO):
    search: Optional[str]
    is_recursive: Optional[bool] = False
    is_in_dataset: Optional[bool]
    item_types: List[StorageItemType]
    include_org_access: Optional[bool] = None
    order: FoldersSortBy
    desc: bool
    page_token: Optional[str]
    page_size: int
    sign_urls: bool


class ListFoldersParams(BaseDTO):
    search: Optional[str] = None
    is_recursive: Optional[bool] = False
    dataset_synced: Optional[bool] = None
    include_org_access: Optional[bool] = None
    order: FoldersSortBy = FoldersSortBy.NAME
    desc: bool = False
    page_token: Optional[str] = None
    page_size: int = 100


class PatchItemPayload(BaseDTO):
    name: Optional[str] = None
    description: Optional[str] = None
    client_metadata: Optional[dict] = None


class PatchFolderPayload(BaseDTO):
    name: Optional[str] = None
    description: Optional[str] = None
    client_metadata: Optional[dict] = None


class PatchFoldersBulkPayload(BaseDTO):
    folder_patches: Dict[str, PatchFolderPayload]


class StorageFolderSummary(BaseDTO):
    files: int
    folders: int
    videos: int
    images: int
    image_groups: int
    image_sequences: int
    dicom_files: int
    dicom_series: int
    niftis: int
    audios: int
    pdfs: int
    plain_texts: int
    tombstones: int
    upload_jobs: int
    total_size: float


class ItemShortInfo(BaseDTO):
    uuid: UUID
    name: str
    parent_uuid: UUID
    parent_name: str
    item_type: StorageItemType


class DatasetShortInfo(BaseDTO):
    dataset_hash: str
    backing_folder_uuid: Optional[UUID]
    title: str
    data_hashes: List[UUID]
    data_units_created_at: datetime


class StorageItemSummary(BaseDTO):
    """A summary of item usage in the system"""

    frame_in_groups: int
    """A number of group items (DICOM_SERIES, IMAGE_GROUP, IMAGE_SEQUENCE) that contain this item"""
    accessible_group_items: List[ItemShortInfo]
    """List of group items that contain this item (only those that the user has access to,
    so the length of this list can be less than `frame_in_groups`)"""
    used_in_datasets: int
    """A number of datasets that contain this item as a `DataRow`"""
    accessible_datasets: List[DatasetShortInfo]
    """List of datasets that contain this item as a `DataRow` (only those that the user has access to, so
    the length of this list can be less than `used_in_datasets`)"""


class DeleteItemsParams(BaseDTO):
    remove_unused_frames: bool


class DeleteItemsPayload(BaseDTO):
    child_uuids: List[UUID]
    remove_unused_frames: bool


class DeleteItemsResponse(BaseDTO):
    removed_items_count: int
    removed_folders_count: int


class MoveItemsPayload(BaseDTO):
    item_uuids: List[UUID]
    new_parent_uuid: UUID
    allow_synced_dataset_move: bool = False


class MoveFoldersPayload(BaseDTO):
    folder_uuids: List[UUID]
    new_parent_uuid: Optional[UUID]


class GetItemParams(BaseDTO):
    sign_url: bool


class GetChildItemsParams(BaseDTO):
    sign_urls: bool = False


class GetItemsBulkPayload(BaseDTO):
    item_uuids: List[UUID]
    sign_urls: bool = False


class PatchItemsBulkPayload(BaseDTO):
    item_patches: Dict[str, PatchItemPayload]


@dataclass
class BundledPatchItemPayload:
    item_patches: Dict[str, PatchItemPayload]

    def add(self, other: BundledPatchItemPayload) -> BundledPatchItemPayload:
        self.item_patches.update(other.item_patches)
        return self


@dataclass
class BundledPatchFolderPayload:
    folder_patches: Dict[str, PatchFolderPayload]

    def add(self, other: BundledPatchFolderPayload) -> BundledPatchFolderPayload:
        self.folder_patches.update(other.folder_patches)
        return self


class ReencodeVideoItemsRequest(BaseDTO):
    storage_items: List[UUID]
    process_title: str
    force_full_reencoding: bool


class JobStatus(CamelStrEnum):
    SUBMITTED = auto()
    DONE = auto()
    ERROR = auto()


class ReencodeVideoItemsResponse(BaseDTO):
    status: JobStatus
    result: Optional[Union[list, dict]]


class StorageItemsMigratePayload(BaseDTO):
    urls_map: Optional[Dict[str, Optional[str]]]
    items_map: Optional[Dict[UUID, Optional[str]]]
    from_integration_hash: Optional[UUID] = None
    to_integration_hash: Optional[UUID] = None
    validate_access: bool = False
    skip_missing: bool = False


class AddDataToFolderJobCancelResponse(BaseDTO):
    units_cancelled_count: int



import base64
import hashlib
from datetime import datetime
from typing import Dict, Union

from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from requests import PreparedRequest

_SIGNATURE_LABEL = "encord-sig"

"""
This file implements request signing according to the following RFC draft:
https://datatracker.ietf.org/doc/html/draft-ietf-httpbis-message-signatures
"""


def _sfv_str(key: str, value: Union[str, bytes]) -> str:
    if isinstance(value, bytes):
        return f"{key}=:{base64.b64encode(value).decode('ascii')}:"
    return f"{key}={value}"


def _sfv_value(value: Union[int, str]) -> str:
    if isinstance(value, int):
        return str(value)
    return f'"{value}"'


def _request_body_bytes(request: PreparedRequest) -> bytes:
    if isinstance(request.body, str):
        return request.body.encode("utf-8")
    elif isinstance(request.body, bytes):
        return request.body
    else:
        return b""


def sign_request(request: PreparedRequest, key_id: str, private_key: Ed25519PrivateKey) -> PreparedRequest:
    assert request.method is not None

    content_digest = _sfv_str("sha-256", hashlib.sha256(_request_body_bytes(request)).digest())
    # Moving 'created' time to be a bit in the past, since sometimes requests are failing
    # due to the clock skew.
    # This is to be fixed on server side and removed from here.
    signature_params: Dict[str, Union[str, int]] = {
        "created": int(datetime.now().timestamp()) - 30,
        "keyid": key_id,
        "alg": "ed25519",
    }

    signature_elements = {
        "@method": request.method.upper(),
        "@request-target": request.path_url,
        "content-digest": content_digest,
    }

    covered_elements = [f'"{element_id}"' for element_id in signature_elements]
    signature_elements_pairs = [f"{k}={_sfv_value(v)}" for k, v in signature_params.items()]
    sig_params_serialised = ";".join([f"({' '.join(covered_elements)})"] + signature_elements_pairs)

    signature_elements["@signature-params"] = sig_params_serialised

    signature_base = "\n".join(f'"{k}": {v}' for k, v in signature_elements.items())
    signature = private_key.sign(signature_base.encode())

    request.headers["Content-Digest"] = content_digest
    request.headers["Signature-Input"] = _sfv_str(_SIGNATURE_LABEL, sig_params_serialised)
    request.headers["Signature"] = _sfv_str(_SIGNATURE_LABEL, signature)

    return request



"""---
title: "Objects - Ontology Structure"
slug: "sdk-ref-objects-ont-structure"
hidden: false
metadata:
  title: "Objects - Ontology Structure"
  description: "Encord SDK Objects - Ontology Structure."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Type, cast
from uuid import uuid4

from encord.exceptions import OntologyError
from encord.objects.classification import Classification
from encord.objects.common import Shape
from encord.objects.constants import AVAILABLE_COLORS
from encord.objects.ontology_element import (
    OntologyElement,
    OntologyElementT,
    OntologyNestedElement,
    _assert_singular_result_list,
    _get_element_by_hash,
)
from encord.objects.ontology_object import Object
from encord.objects.skeleton_template import SkeletonTemplate
from encord.objects.utils import checked_cast, does_type_match


@dataclass
class OntologyStructure:
    objects: List[Object] = field(default_factory=list)
    classifications: List[Classification] = field(default_factory=list)
    skeleton_templates: Dict[str, SkeletonTemplate] = field(default_factory=dict)

    def get_child_by_hash(
        self,
        feature_node_hash: str,
        type_: Optional[Type[OntologyElementT]] = None,
    ) -> OntologyElementT:
        """Returns the first child node of this ontology tree node with the matching feature node hash. If there is
        more than one child with the same feature node hash in the ontology tree node, then the ontology would be in
        an invalid state. Throws if nothing is found or if the type is not matched.

        Args:
            feature_node_hash: The feature_node_hash of the child node to search for in the ontology.
            type_: The expected type of the item. If the found child does not match the type, an error will be thrown.

        Raises:
            OntologyError: If the item with the specified feature_node_hash is not found or if the type does not match.
        """
        for object_ in self.objects:
            if object_.feature_node_hash == feature_node_hash:
                return checked_cast(object_, type_)

            found_item = _get_element_by_hash(feature_node_hash, object_.attributes)
            if found_item is not None:
                return checked_cast(found_item, type_)

        for classification in self.classifications:
            if classification.feature_node_hash == feature_node_hash:
                return checked_cast(classification, type_)
            found_item = _get_element_by_hash(feature_node_hash, classification.attributes)
            if found_item is not None:
                return checked_cast(found_item, type_)

        raise OntologyError(f"Item not found: can't find an item with a hash {feature_node_hash} in the ontology.")

    def get_child_by_title(
        self,
        title: str,
        type_: Optional[Type[OntologyElementT]] = None,
    ) -> OntologyElementT:
        """Returns a child node of this ontology tree node with the matching title and matching type if specified. If more
        than one child in this Object have the same title, then an error will be thrown. If no item is found, an error
        will be thrown as well.

        Args:
            title: The exact title of the child node to search for in the ontology.
            type_: The expected type of the child node. Only a node that matches this type will be returned.

        Raises:
            OntologyError: If no child node with the specified title and type is found, or if multiple matches are found.
        """
        found_items = self.get_children_by_title(title, type_)
        _assert_singular_result_list(found_items, title, type_)
        return found_items[0]

    def get_children_by_title(
        self,
        title: str,
        type_: Optional[Type[OntologyElementT]] = None,
    ) -> List[OntologyElementT]:
        """Returns all the child nodes of this ontology tree node with the matching title and matching type if specified.
        Title in ontologies do not need to be unique, however, we recommend unique titles when creating ontologies.

        Args:
            title: The exact title of the child node to search for in the ontology.
            type_: The expected type of the item. Only nodes that match this type will be returned.

        Returns:
            List[OntologyElementT]: A list of child nodes with the matching title and type.
        """
        ret: List[OntologyElement] = []
        for object_ in self.objects:
            if object_.title == title and does_type_match(object_, type_):
                ret.append(object_)

            if type_ is None or issubclass(type_, OntologyNestedElement):
                found_items = object_.get_children_by_title(title, type_=type_)
                ret.extend(found_items)

        for classification in self.classifications:
            if classification.title == title and does_type_match(classification, type_):
                ret.append(classification)

            if type_ is None or issubclass(type_, OntologyNestedElement):
                found_items = classification.get_children_by_title(title, type_=type_)
                ret.extend(found_items)

        # type checks in the code above guarantee the type conformity of the return value
        # but there is no obvious way to tell that to mypy, so just casting here for now
        return cast(List[OntologyElementT], ret)

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> OntologyStructure:
        """Create an OntologyStructure from a dictionary.

        Args:
            d: A JSON blob of an "ontology structure" (e.g. from Encord web app)

        Returns:
            OntologyStructure: The created OntologyStructure object.

        Raises:
            KeyError: If the dict is missing a required field.
        """
        objects_ret = [Object.from_dict(object_dict) for object_dict in d["objects"]]
        classifications_ret = [
            Classification.from_dict(classification_dict) for classification_dict in d["classifications"]
        ]
        skeleton_templates = [
            SkeletonTemplate.from_dict(skeleton_template_dict["template"])
            for skeleton_template_dict in d.get("skeleton_templates", [])
        ]
        dict_skeleton_templates = {template.name: template for template in skeleton_templates}
        return OntologyStructure(
            objects=objects_ret, classifications=classifications_ret, skeleton_templates=dict_skeleton_templates
        )

    def to_dict(self) -> Dict[str, List[Dict[str, Any]]]:
        """Convert the OntologyStructure to a dictionary.

        Returns:
            Dict[str, List[Dict[str, Any]]]: The dictionary representation of the ontology.
        """
        ret: Dict[str, List[Dict[str, Any]]] = dict()
        ontology_objects: List[Dict[str, Any]] = list()
        ret["objects"] = ontology_objects
        for ontology_object in self.objects:
            ontology_objects.append(ontology_object.to_dict())

        ontology_classifications: List[Dict[str, Any]] = list()
        ret["classifications"] = ontology_classifications
        for ontology_classification in self.classifications:
            ontology_classifications.append(ontology_classification.to_dict())
        skeleton_templates: List[Dict[str, Any]] = list()
        skeleton_templates = [{"template": template.to_dict()} for template in self.skeleton_templates.values()]
        ret["skeleton_templates"] = skeleton_templates
        return ret

    def add_object(
        self,
        name: str,
        shape: Shape,
        uid: Optional[int] = None,
        color: Optional[str] = None,
        feature_node_hash: Optional[str] = None,
    ) -> Object:
        """Adds an object class definition to the structure.

        Args:
            name: The user-visible name of the object.
            shape: The kind of object (bounding box, polygon, etc). See :py:class:`encord.objects.common.Shape` enum for possible values.
            uid: Integer identifier of the object. Normally auto-generated; omit this unless the aim is to create an exact clone of existing structure.
            color: The color of the object in the label editor. Normally auto-assigned, should be in '#1A2B3F' syntax.
            feature_node_hash: Global identifier of the object. Normally auto-generated; omit this unless the aim is to create an exact clone of existing structure.

        Returns:
            Object: The created object class that can be further customized with attributes.

        Raises:
            ValueError: If a duplicate uid or feature_node_hash is provided.
        """
        if uid is None:
            if self.objects:
                uid = max([obj.uid for obj in self.objects]) + 1
            else:
                uid = 1
        else:
            if any([obj.uid == uid for obj in self.objects]):
                raise ValueError(f"Duplicate uid '{uid}'")

        if color is None:
            color_index = 0
            if self.objects:
                try:
                    color_index = AVAILABLE_COLORS.index(self.objects[-1].color) + 1
                    if color_index >= len(AVAILABLE_COLORS):
                        color_index = 0
                except ValueError:
                    pass
            color = AVAILABLE_COLORS[color_index]

        if feature_node_hash is None:
            feature_node_hash = str(uuid4())[:8]

        if any([obj.feature_node_hash == feature_node_hash for obj in self.objects]):
            raise ValueError(f"Duplicate feature_node_hash '{feature_node_hash}'")

        obj = Object(uid=uid, name=name, color=color, shape=shape, feature_node_hash=feature_node_hash)
        self.objects.append(obj)
        return obj

    def add_classification(
        self,
        uid: Optional[int] = None,
        feature_node_hash: Optional[str] = None,
    ) -> Classification:
        """Adds a classification definition to the ontology.

        Args:
            uid: Integer identifier of the object. Normally auto-generated; omit this unless the aim is to create an exact clone of existing structure.
            feature_node_hash: Global identifier of the object. Normally auto-generated; omit this unless the aim is to create an exact clone of existing structure.

        Returns:
            Classification: The created classification node. Note that classification attribute should be further specified by calling its `add_attribute()` method.

        Raises:
            ValueError: If a duplicate uid or feature_node_hash is provided.
        """
        if uid is None:
            if self.classifications:
                uid = max([cls.uid for cls in self.classifications]) + 1
            else:
                uid = 1
        else:
            if any([cls.uid == uid for cls in self.classifications]):
                raise ValueError(f"Duplicate uid '{uid}'")

        if feature_node_hash is None:
            feature_node_hash = str(uuid4())[:8]

        if any([cls.feature_node_hash == feature_node_hash for cls in self.classifications]):
            raise ValueError(f"Duplicate feature_node_hash '{feature_node_hash}'")

        cls = Classification(uid=uid, feature_node_hash=feature_node_hash, attributes=list())
        self.classifications.append(cls)
        return cls

    def add_skeleton_template(
        self,
        skeleton_template: SkeletonTemplate,
        feature_node_hash: Optional[str] = None,
    ) -> None:
        """Adds a skeleton template to the ontology structure.

        Args:
            skeleton_template: The SkeletonTemplate object to be added.
            feature_node_hash: Global identifier of the skeleton template. Normally auto-generated; omit this unless the aim is to create an exact clone of existing structure.

        Raises:
            ValueError: If a skeleton template with the same name already exists in the ontology.
        """
        if feature_node_hash is None:
            feature_node_hash = str(uuid4())[:8]
        skeleton_template.feature_node_hash = feature_node_hash
        if skeleton_template.name in self.skeleton_templates:
            raise ValueError("Already a template with this name associated to this ontology")
        self.skeleton_templates[skeleton_template.name] = skeleton_template



"""---
title: "Objects - Coordinates"
slug: "sdk-ref-objects-coordinates"
hidden: false
metadata:
  title: "Objects - Coordinates"
  description: "Encord SDK Objects - Coordinates."
category: "64e481b57b6027003f20aaa0"
---
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Type, Union

from encord.exceptions import LabelRowError
from encord.objects.bitmask import BitmaskCoordinates
from encord.objects.common import Shape
from encord.objects.frames import Ranges
from encord.objects.html_node import HtmlRange
from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO


@dataclass(frozen=True)
class BoundingBoxCoordinates:
    """Represents bounding box coordinates, where all values are percentages relative to the total image size.

    Attributes:
        height (float): The height of the bounding box.
        width (float): The width of the bounding box.
        top_left_x (float): The x-coordinate of the top-left corner.
        top_left_y (float): The y-coordinate of the top-left corner.
    """

    height: float
    width: float
    top_left_x: float
    top_left_y: float

    @staticmethod
    def from_dict(d: dict) -> BoundingBoxCoordinates:
        """Create a BoundingBoxCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing bounding box information.

        Returns:
            BoundingBoxCoordinates: An instance of BoundingBoxCoordinates.
        """
        bounding_box_dict = d["boundingBox"]
        return BoundingBoxCoordinates(
            height=bounding_box_dict["h"],
            width=bounding_box_dict["w"],
            top_left_x=bounding_box_dict["x"],
            top_left_y=bounding_box_dict["y"],
        )

    def to_dict(self) -> dict:
        """Convert the BoundingBoxCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the bounding box coordinates.
        """
        return {
            "h": self.height,
            "w": self.width,
            "x": self.top_left_x,
            "y": self.top_left_y,
        }


@dataclass(frozen=True)
class RotatableBoundingBoxCoordinates:
    """Represents rotatable bounding box coordinates, where all values are percentages relative to the total image size.

    Attributes:
        height (float): The height of the bounding box.
        width (float): The width of the bounding box.
        top_left_x (float): The x-coordinate of the top-left corner.
        top_left_y (float): The y-coordinate of the top-left corner.
        theta (float): The angle of rotation originating at the center of the box.
    """

    height: float
    width: float
    top_left_x: float
    top_left_y: float
    theta: float  # angle of rotation originating at center of box

    @staticmethod
    def from_dict(d: dict) -> RotatableBoundingBoxCoordinates:
        """Create a RotatableBoundingBoxCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing rotatable bounding box information.

        Returns:
            RotatableBoundingBoxCoordinates: An instance of RotatableBoundingBoxCoordinates.
        """
        rotatable_bounding_box_dict = d["rotatableBoundingBox"]
        return RotatableBoundingBoxCoordinates(
            height=rotatable_bounding_box_dict["h"],
            width=rotatable_bounding_box_dict["w"],
            top_left_x=rotatable_bounding_box_dict["x"],
            top_left_y=rotatable_bounding_box_dict["y"],
            theta=rotatable_bounding_box_dict["theta"],
        )

    def to_dict(self) -> dict:
        """Convert the RotatableBoundingBoxCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the rotatable bounding box coordinates.
        """
        return {
            "h": self.height,
            "w": self.width,
            "x": self.top_left_x,
            "y": self.top_left_y,
            "theta": self.theta,
        }


@dataclass(frozen=True)
class PointCoordinate:
    """Represents a point coordinate, where all coordinates are a percentage relative to the total image size.

    Attributes:
        x (float): The x-coordinate of the point.
        y (float): The y-coordinate of the point.
    """

    x: float
    y: float

    @staticmethod
    def from_dict(d: dict) -> PointCoordinate:
        """Create a PointCoordinate instance from a dictionary.

        Args:
            d (dict): A dictionary containing point coordinate information.

        Returns:
            PointCoordinate: An instance of PointCoordinate.
        """
        first_item = d["point"]["0"]
        return PointCoordinate(
            x=first_item["x"],
            y=first_item["y"],
        )

    def to_dict(self) -> dict:
        """Convert the PointCoordinate instance to a dictionary.

        Returns:
            dict: A dictionary representation of the point coordinate.
        """
        return {"0": {"x": self.x, "y": self.y}}


class PolygonCoordsToDict(str, Enum):
    single_polygon = "single_polygon"
    multiple_polygons = "multiple_polygons"


class PolygonCoordinates:
    """Represents polygon coordinates"""

    def __init__(
        self, values: list[PointCoordinate] | None = None, polygons: list[list[list[PointCoordinate]]] | None = None
    ):
        """
        Args:
            values (List[PointCoordinate]): A list of PointCoordinate objects defining the polygon.
            polygons (List[List[List[PointCoordinate]]]): A list of polygons, where each polygon is a list of contours, where each contour is a list of points.
        """

        if not values and not polygons:
            raise LabelRowError("Either `values` or `polygons` must be provided")
        elif values and not polygons:
            self._values = values
            self._polygons = [[list(values)]]
        elif polygons and not values:
            self._polygons = polygons
            self._values = [point for point in self._polygons[0][0]]
        elif polygons and values:
            if polygons[0][0] != values:
                raise LabelRowError("`values` and `polygons` are not consistent")
            self._values = values
            self._polygons = polygons

    @property
    def values(self) -> list[PointCoordinate]:
        return self._values

    @property
    def polygons(self) -> list[list[list[PointCoordinate]]]:
        return self._polygons

    @staticmethod
    def from_dict(d: dict) -> "PolygonCoordinates":
        """Create a PolygonCoordinates instance from a dictionary.

        Supports both legacy format (single polygon with one contour) and new complex format.

        Args:
            d (dict): A dictionary containing polygon coordinates information.
                Legacy format: {"polygon": {str(idx): {"x": x, "y": y}} or {"polygon": [{"x": x, "y": y}]}
                New format: {"polygons": [[[{"x": x, "y": y}]]]}

        Returns:
            PolygonCoordinates: An instance of PolygonCoordinates.
        """
        polygon_dict = d.get("polygon")
        values: List[PointCoordinate] = []

        if isinstance(polygon_dict, dict):
            sorted_dict_value_tuples = sorted((int(key), value) for key, value in polygon_dict.items())
            sorted_dict_values = [item[1] for item in sorted_dict_value_tuples]
        elif isinstance(polygon_dict, list):
            sorted_dict_values = list(polygon_dict)
        elif not polygon_dict:  # Empty dict case
            sorted_dict_values = []
        else:
            raise LabelRowError(f"Invalid format for polygon coordinates: {polygon_dict}")

        for value in sorted_dict_values:
            point_coordinate = PointCoordinate(
                x=value["x"],
                y=value["y"],
            )
            values.append(point_coordinate)

        # Parse new format if present
        polygons = []
        for polygon in d.get("polygons", []):
            contours = [flat_to_pnt_coordinates(contour) for contour in polygon]
            polygons.append(contours)

        return PolygonCoordinates(values=values, polygons=polygons)

    def to_dict(
        self, kind: PolygonCoordsToDict | str = PolygonCoordsToDict.single_polygon
    ) -> dict | list[list[list[float]]]:
        """Convert the PolygonCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the polygon coordinates.
        """
        if kind == PolygonCoordsToDict.single_polygon:
            return {str(idx): {"x": value.x, "y": value.y} for idx, value in enumerate(self._values)}
        elif kind == PolygonCoordsToDict.multiple_polygons:
            return [[pnt_coordinates_to_flat(contour) for contour in polygon] for polygon in self._polygons]
        else:
            raise LabelRowError(f"Invalid argument: {kind}")


def flat_to_pnt_coordinates(ring: list[float]) -> list[PointCoordinate]:
    return [PointCoordinate(x=ring[idx], y=ring[idx + 1]) for idx in range(0, len(ring), 2)]


def pnt_coordinates_to_flat(coordinates: list[PointCoordinate]) -> list[float]:
    return [coord for point in coordinates for coord in [point.x, point.y]]


@dataclass(frozen=True)
class PolylineCoordinates:
    """Represents polyline coordinates as a list of point coordinates.

    Attributes:
        values (List[PointCoordinate]): A list of PointCoordinate objects defining the polyline.
    """

    values: List[PointCoordinate]

    @staticmethod
    def from_dict(d: dict) -> PolylineCoordinates:
        """Create a PolylineCoordinates instance from a dictionary.

        Args:
            d (dict): A dictionary containing polyline coordinates information.

        Returns:
            PolylineCoordinates: An instance of PolylineCoordinates.
        """
        polyline = d["polyline"]
        values: List[PointCoordinate] = []

        if isinstance(polyline, dict):
            sorted_dict_value_tuples = sorted((int(key), value) for key, value in polyline.items())
            sorted_dict_values = [item[1] for item in sorted_dict_value_tuples]
        elif isinstance(polyline, list):
            sorted_dict_values = list(polyline)
        else:
            raise LabelRowError(f"Invalid format for polyline coordinates: {polyline}")

        for value in sorted_dict_values:
            point_coordinate = PointCoordinate(
                x=value["x"],
                y=value["y"],
            )
            values.append(point_coordinate)

        return PolylineCoordinates(values=values)

    def to_dict(self) -> dict:
        """Convert the PolylineCoordinates instance to a dictionary.

        Returns:
            dict: A dictionary representation of the polyline coordinates.
        """
        return {str(idx): {"x": value.x, "y": value.y} for idx, value in enumerate(self.values)}


class Visibility(CamelStrEnum):
    """An enumeration to represent the visibility state of an item.

    Attributes:
        VISIBLE: The item is visible within the frame.
        INVISIBLE: The item is outside the frame and thus invisible.
        OCCLUDED: The item is within the frame but occluded by something else.
    """

    VISIBLE = auto()
    INVISIBLE = auto()
    OCCLUDED = auto()


class SkeletonCoordinate(BaseDTO):
    """Represents a coordinate for a skeleton structure in an image.

    Attributes:
        x (float): The x-coordinate of the skeleton point.
        y (float): The y-coordinate of the skeleton point.
        name (str): The name of the skeleton point.
        color (Optional[str]): The color associated with the skeleton point.
        feature_hash (Optional[str]): A unique hash for the feature.
        value (Optional[str]): An optional value associated with the skeleton point.
        visibility (Optional[Visibility]): The visibility state of the skeleton point.
    """

    x: float
    y: float
    name: str
    color: Optional[str] = None
    feature_hash: Optional[str] = None
    value: Optional[str] = None

    visibility: Optional[Visibility] = None


class SkeletonCoordinates(BaseDTO):
    """Represents a collection of skeleton coordinates.

    Attributes:
        values (List[SkeletonCoordinate]): A list of SkeletonCoordinate objects.
        name (str): The name of the skeleton structure.
    """

    values: List[SkeletonCoordinate]
    name: str

    def to_dict(self, by_alias=True, exclude_none=True) -> Dict[str, Any]:
        """Convert the SkeletonCoordinates instance to a dictionary.

        Args:
            by_alias (bool): Whether to use alias for the field names.
            exclude_none (bool): Whether to exclude fields with None values.

        Returns:
            Dict[str, Any]: A dictionary representation of the skeleton coordinates.
        """
        return {str(i): x.to_dict() for i, x in enumerate(self.values)}


class AudioCoordinates(BaseDTO):
    """Represents coordinates for an audio file

    Attributes:
        range (Ranges): Ranges in milliseconds for audio files
    """

    range: Ranges

    def __post_init__(self):
        if len(self.range) == 0:
            raise ValueError("Range list must contain at least one range.")


class TextCoordinates(BaseDTO):
    """Represents coordinates for a text file

    Attributes:
        range (Ranges): Ranges of chars for simple text files
    """

    range: Ranges


class HtmlCoordinates(BaseDTO):
    """Represents coordinates for a html file

    Attributes:
        range (List[HtmlRange]): A list of HtmlRange objects
    """

    range: List[HtmlRange]


NON_GEOMETRIC_COORDINATES = {AudioCoordinates, TextCoordinates, HtmlCoordinates}


Coordinates = Union[
    AudioCoordinates,
    TextCoordinates,
    HtmlCoordinates,
    BoundingBoxCoordinates,
    RotatableBoundingBoxCoordinates,
    PointCoordinate,
    PolygonCoordinates,
    PolylineCoordinates,
    SkeletonCoordinates,
    BitmaskCoordinates,
]

ACCEPTABLE_COORDINATES_FOR_ONTOLOGY_ITEMS: Dict[Shape, List[Type[Coordinates]]] = {
    Shape.BOUNDING_BOX: [BoundingBoxCoordinates],
    Shape.ROTATABLE_BOUNDING_BOX: [RotatableBoundingBoxCoordinates],
    Shape.POINT: [PointCoordinate],
    Shape.POLYGON: [PolygonCoordinates],
    Shape.POLYLINE: [PolylineCoordinates],
    Shape.SKELETON: [SkeletonCoordinates],
    Shape.BITMASK: [BitmaskCoordinates],
    Shape.AUDIO: [AudioCoordinates],
    Shape.TEXT: [TextCoordinates, HtmlCoordinates],
}



from typing import List, Sequence, Tuple

import numpy as np

# Importing python implementations of functions that not have numpy implementation
from .bitmask_operations import _rle_to_mask, _rle_to_string, _string_to_rle


def _mask_to_rle(mask: bytes) -> List[int]:
    """COCO-compatible raw bitmask to COCO-compatible RLE"""
    mask_buffer = np.frombuffer(mask, dtype=np.bool_)
    changes = np.diff(mask_buffer, prepend=mask_buffer[0], append=mask_buffer[-1])
    change_indices = np.flatnonzero(changes != 0)
    run_lengths = np.diff(np.concatenate(([0], change_indices, [len(mask_buffer)])))
    return run_lengths.tolist()


def serialise_bitmask(bitmask: bytes) -> str:
    rle = _mask_to_rle(bitmask)
    return _rle_to_string(rle)


def deserialise_bitmask(serialised_bitmask: str, length: int) -> bytes:
    rle = _string_to_rle(serialised_bitmask)
    return _rle_to_mask(rle, length)


def transpose_bytearray(byte_data: bytes, shape: Tuple[int, int]) -> bytes:
    np_byte_data = np.frombuffer(byte_data, dtype=np.int8).reshape(shape)
    return bytearray(np_byte_data.T.tobytes())



"""---
title: "Objects - DICOM Metadata"
slug: "sdk-ref-objects-dicom-metadata"
hidden: false
metadata:
  title: "Objects - DICOM Metadata"
  description: "Encord SDK Objects - DICOM Metadata."
category: "64e481b57b6027003f20aaa0"
---
"""

from typing import Optional

from encord.orm.base_dto import BaseDTO


class DICOMSeriesMetadata(BaseDTO):
    """Metadata for a DICOM series.

    Attributes:
        patient_id (Optional[str]): The ID of the patient. This attribute is optional.
        study_uid (str): The unique identifier for the study.
        series_uid (str): The unique identifier for the series.
    """

    patient_id: Optional[str]
    study_uid: str
    series_uid: str


class DICOMSliceMetadata(BaseDTO):
    """Metadata for a slice in a DICOM series.

    Attributes:
        dicom_instance_uid (str): The unique identifier for the DICOM instance.
        multiframe_frame_number (Optional[int]): The frame number if the DICOM instance is a multiframe image. This attribute is optional.
        file_uri (str): The URI to the file containing the slice.
        width (int): The width of the slice in pixels.
        height (int): The height of the slice in pixels.
    """

    dicom_instance_uid: str
    multiframe_frame_number: Optional[int]
    file_uri: str
    width: int
    height: int



"""``encord.client`` provides a simple Python client that allows you
to query project resources through the Encord API.

Here is a simple example for instantiating the client for a project
and obtaining project info:

.. test_blurb2.py code::

    from encord.client import EncordClient

    client = EncordClient.initialize('YourProjectID', 'YourAPIKey')
    client.get_project()

Returns:
        Project: A project record instance. See Project ORM for details.

"""

from __future__ import annotations

import dataclasses
import json
import logging
import os.path
import time
import typing
import uuid
from datetime import datetime
from math import ceil
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple, Union, cast
from uuid import UUID

import requests

import encord.exceptions
from encord.common.deprecated import deprecated
from encord.configs import Config
from encord.constants.enums import DataType
from encord.constants.string_constants import (
    FITTED_BOUNDING_BOX,
    INTERPOLATION,
)
from encord.http.querier import Querier
from encord.http.utils import (
    CloudUploadSettings,
    upload_to_signed_url_list,
)
from encord.http.v2.api_client import ApiClient
from encord.http.v2.payloads import Page
from encord.orm.active import ActiveProjectImportPayload, ActiveProjectMode
from encord.orm.analytics import CollaboratorTimer, CollaboratorTimerParams, TimeSpent, TimeSpentParams
from encord.orm.bearer_request import BearerTokenResponse
from encord.orm.cloud_integration import CloudIntegration, GetCloudIntegrationsResponse
from encord.orm.dataset import (
    DEFAULT_DATASET_ACCESS_SETTINGS,
    AddPrivateDataResponse,
    DataLinkDuplicatesBehavior,
    DataRow,
    DataRows,
    DatasetAccessSettings,
    DatasetData,
    DatasetDataLongPolling,
    DatasetLinkItems,
    DatasetUser,
    DatasetUserRole,
    DatasetUsers,
    DicomSeries,
    Image,
    ImageGroup,
    ImageGroupOCR,
    Images,
    LongPollingStatus,
    ReEncodeVideoTask,
    Video,
)
from encord.orm.dataset import Dataset as OrmDataset
from encord.orm.group import (
    AddDatasetGroupsPayload,
    AddProjectGroupsPayload,
    DatasetGroup,
    ProjectGroup,
    RemoveGroupsParams,
)
from encord.orm.label_log import LabelLog, LabelLogParams
from encord.orm.label_row import (
    AnnotationTaskStatus,
    LabelRow,
    LabelRowMetadata,
    LabelStatus,
    LabelValidationState,
    Review,
    ShadowDataState,
)
from encord.orm.labeling_algorithm import (
    BoundingBoxFittingParams,
    LabelingAlgorithm,
    ObjectInterpolationParams,
)
from encord.orm.project import (
    CopyDatasetAction,
    CopyDatasetOptions,
    CopyLabelsOptions,
    CopyProjectPayload,
    ProjectCopy,
    ProjectCopyOptions,
    ProjectDataset,
    ProjectUsers,
    TaskPriorityParams,
)
from encord.orm.project import Project as OrmProject
from encord.orm.project import ProjectDTO as ProjectOrmV2
from encord.orm.storage import (
    DatasetDataLongPollingParams,
    DataUploadDicomSeries,
    DataUploadDicomSeriesDicomFile,
    DataUploadImage,
    DataUploadImageGroup,
    DataUploadImageGroupImage,
    DataUploadItems,
    DataUploadVideo,
)
from encord.orm.workflow import (
    LabelWorkflowGraphNode,
    LabelWorkflowGraphNodePayload,
    WorkflowAction,
)
from encord.project_ontology.classification_type import ClassificationType
from encord.project_ontology.object_type import ObjectShape
from encord.project_ontology.ontology import Ontology as LegacyOntology
from encord.utilities.client_utilities import optional_datetime_to_iso_str, optional_set_to_list
from encord.utilities.project_user import ProjectUser, ProjectUserRole

LONG_POLLING_RESPONSE_RETRY_N = 3
LONG_POLLING_SLEEP_ON_FAILURE_SECONDS = 10
LONG_POLLING_MAX_REQUEST_TIME_SECONDS = 60
LONG_POLLING_MAX_REQUEST_SINGLE_ITEM_TIME_SECONDS = 10

logger = logging.getLogger(__name__)


class EncordClient:
    """Encord client. Allows you to query db items associated
    with a project (e.g. label rows, datasets).
    """

    def __init__(self, querier: Querier, config: Config, api_client: ApiClient):
        self._querier = querier
        self._config = config
        self._api_client = api_client

    @deprecated(version="0.1.154", alternative="EncordUserClient.get_cloud_integrations")
    def get_cloud_integrations(self) -> List[CloudIntegration]:
        return [
            CloudIntegration(
                id=str(x.integration_uuid),
                title=x.title,
            )
            for x in self._api_client.get(
                "cloud-integrations",
                params=None,
                result_type=GetCloudIntegrationsResponse,
            ).result
        ]

    def get_bearer_token(self) -> BearerTokenResponse:
        return self._api_client.get("user/bearer-token", None, result_type=BearerTokenResponse)


class EncordClientDataset(EncordClient):
    """DEPRECATED - prefer using the :class:`encord.dataset.Dataset` instead"""

    def __init__(
        self,
        querier: Querier,
        config: Config,
        dataset_access_settings: DatasetAccessSettings = DEFAULT_DATASET_ACCESS_SETTINGS,
        api_client: Optional[ApiClient] = None,
    ):
        if api_client is None:
            raise ValueError("api_client is None")

        super().__init__(querier, config, api_client)
        self._dataset_access_settings = dataset_access_settings

    def get_dataset(self) -> OrmDataset:
        """Retrieve dataset info (pointers to data, labels).

        Returns:
            OrmDataset: A dataset record instance.

        Raises:
            AuthorisationError: If the dataset API key is invalid.
            ResourceNotFoundError: If no dataset exists by the specified dataset EntityId.
            UnknownError: If an error occurs while retrieving the dataset.
        """
        res = self._querier.basic_getter(
            OrmDataset,
            payload={
                "dataset_access_settings": dataclasses.asdict(self._dataset_access_settings),
            },
        )

        for row in res.data_rows:
            row["_querier"] = self._querier
        return res

    def list_data_rows(
        self,
        title_eq: Optional[str] = None,
        title_like: Optional[str] = None,
        created_before: Optional[Union[str, datetime]] = None,
        created_after: Optional[Union[str, datetime]] = None,
        data_types: Optional[List[DataType]] = None,
        data_hashes: Optional[List[str]] = None,
    ) -> List[DataRow]:
        """Retrieve dataset rows (pointers to data, labels).

        Args:
            title_eq: optional exact title row filter
            title_like: optional fuzzy title row filter; SQL syntax
            created_before: optional datetime row filter
            created_after: optional datetime row filter
            data_types: optional data types row filter
            data_hashes: optional list of individual data unit hashes to include

        Returns:
            List[DataRow]: A list of DataRows object that match the filter

        Raises:
            AuthorisationError: If the dataset API key is invalid.
            ResourceNotFoundError: If no dataset exists by the specified dataset EntityId.
            UnknownError: If an error occurs while retrieving the dataset.
        """
        created_before = optional_datetime_to_iso_str("created_before", created_before)
        created_after = optional_datetime_to_iso_str("created_after", created_after)

        data_rows = cast(
            List[DataRow],
            self._querier.get_multiple(
                DataRows,
                payload={
                    "title_eq": title_eq,
                    "title_like": title_like,
                    "created_before": created_before,
                    "created_after": created_after,
                    "data_types": [data_type.to_upper_case_string() for data_type in data_types]
                    if data_types is not None
                    else None,
                    "dataset_access_settings": dataclasses.asdict(self._dataset_access_settings),
                    "data_hashes": data_hashes,
                },
            ),
        )

        for row in data_rows:
            row["_querier"] = self._querier

        return data_rows

    def set_access_settings(self, dataset_access_settings: DatasetAccessSettings) -> None:
        self._dataset_access_settings = dataset_access_settings

    def add_users(self, user_emails: List[str], user_role: DatasetUserRole) -> List[DatasetUser]:
        """This function is documented in :meth:`encord.project.Dataset.add_users`."""
        payload = {"user_emails": user_emails, "user_role": user_role}
        users = self._querier.basic_setter(DatasetUsers, self._querier.resource_id, payload=payload)

        return [DatasetUser.from_dict(user) for user in users]

    def list_groups(self, dataset_hash: uuid.UUID) -> Page[DatasetGroup]:
        return self._api_client.get(f"datasets/{dataset_hash}/groups", params=None, result_type=Page[DatasetGroup])

    def add_groups(self, dataset_hash: str, group_hash: List[uuid.UUID], user_role: DatasetUserRole) -> None:
        payload = AddDatasetGroupsPayload(group_hash_list=group_hash, user_role=user_role)
        self._api_client.post(f"datasets/{dataset_hash}/groups", params=None, payload=payload, result_type=None)

    def remove_groups(self, dataset_hash: uuid.UUID, group_hash: List[uuid.UUID]) -> None:
        params = RemoveGroupsParams(group_hash_list=group_hash)
        self._api_client.delete(f"datasets/{dataset_hash}/groups", params=params, result_type=None)

    def __add_data_to_dataset_get_result(
        self,
        upload_job_id: str,
        timeout_seconds: int = 7 * 24 * 60 * 60,  # 7 days
    ) -> DatasetDataLongPolling:
        failed_requests_count = 0
        polling_start_timestamp = time.perf_counter()

        while True:
            try:
                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                res = self._querier.basic_getter(
                    DatasetDataLongPolling,
                    self._querier.resource_id,
                    payload={
                        "process_hash": upload_job_id,
                        "timeout_seconds": min(
                            polling_available_seconds,
                            LONG_POLLING_MAX_REQUEST_SINGLE_ITEM_TIME_SECONDS,
                        ),
                    },
                )

                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                if polling_available_seconds == 0 or res.status in [
                    LongPollingStatus.DONE,
                    LongPollingStatus.ERROR,
                    LongPollingStatus.CANCELLED,
                ]:
                    return res

                failed_requests_count = 0
            except (requests.exceptions.RequestException, encord.exceptions.RequestException):
                failed_requests_count += 1

                if failed_requests_count >= LONG_POLLING_RESPONSE_RETRY_N:
                    raise

                time.sleep(LONG_POLLING_SLEEP_ON_FAILURE_SECONDS)

    def upload_video(
        self,
        file_path: Union[str, Path],
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> Video:
        """This function is documented in :meth:`encord.dataset.Dataset.upload_video`."""
        signed_url = upload_to_signed_url_list(
            file_paths=[file_path],
            config=self._config,
            querier=self._querier,
            orm_class=Video,
            cloud_upload_settings=cloud_upload_settings,
        )[0]

        upload_job_id = self._querier.basic_setter(
            DatasetDataLongPolling,
            uid=self._querier.resource_id,
            payload=DatasetDataLongPollingParams(
                data_items=DataUploadItems(
                    videos=[
                        DataUploadVideo(
                            object_url=signed_url["file_link"],
                            title=title or signed_url["title"],
                        )
                    ],
                ),
                files=None,
                integration_id=None,
                ignore_errors=False,
                folder_uuid=folder_uuid,
                file_name=None,
            ),
        )["process_hash"]

        res = self.__add_data_to_dataset_get_result(upload_job_id)

        if res.status == LongPollingStatus.DONE:
            if len(res.data_hashes_with_titles) != 1:
                raise encord.exceptions.EncordException(
                    f"An error has occurred during video upload. len({res.data_hashes_with_titles=}) != 1, {upload_job_id=}"
                )

            res_item = res.data_hashes_with_titles[0]
            logger.info(f"Upload complete. {upload_job_id=}")

            return Video(  # Video model types annotations are not correct
                {
                    "data_hash": res_item.data_hash,
                    "title": res_item.title,
                }
            )
        elif res.status == LongPollingStatus.ERROR:
            raise encord.exceptions.EncordException(
                f"An error has occurred during video upload. {res.errors=}, {upload_job_id=}"
            )
        else:
            raise ValueError(f"{res.status=}, this should never happen, {upload_job_id=}")

    def create_image_group(
        self,
        file_paths: Iterable[Union[str, Path]],
        max_workers: Optional[int] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        *,
        create_video: bool = True,
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> List[ImageGroup]:
        """This function is documented in :meth:`encord.dataset.Dataset.create_image_group`."""
        signed_urls = upload_to_signed_url_list(
            file_paths=file_paths,
            config=self._config,
            querier=self._querier,
            orm_class=Images,
            cloud_upload_settings=cloud_upload_settings,
        )

        if not signed_urls:
            raise encord.exceptions.EncordException("All image uploads failed. Image group was not created.")

        upload_job_id = self._querier.basic_setter(
            DatasetDataLongPolling,
            uid=self._querier.resource_id,
            payload=DatasetDataLongPollingParams(
                data_items=DataUploadItems(
                    image_groups=[
                        DataUploadImageGroup(
                            images=[
                                DataUploadImageGroupImage(
                                    url=x["file_link"],
                                    title=x["title"],
                                )
                                for x in signed_urls
                            ],
                            title=title,
                            create_video=create_video,
                            cluster_by_resolution=create_video,  # cluster_by_resolution only if videos are created
                        )
                    ],
                ),
                files=None,
                integration_id=None,
                ignore_errors=False,
                folder_uuid=folder_uuid,
                file_name=None,
            ),
        )["process_hash"]

        res = self.__add_data_to_dataset_get_result(upload_job_id)

        if res.status == LongPollingStatus.DONE:
            if not res.data_hashes_with_titles:
                raise encord.exceptions.EncordException(
                    f"An error has occurred during image group creation. {res.data_hashes_with_titles=} is empty, {upload_job_id=}"
                )

            titles = [x.title for x in res.data_hashes_with_titles]
            logger.info(f"Upload successful! {titles} created, {upload_job_id=}")

            return [
                ImageGroup(
                    {
                        "data_hash": x.data_hash,
                        "title": x.title,
                    }
                )
                for x in res.data_hashes_with_titles
            ]
        elif res.status == LongPollingStatus.ERROR:
            raise encord.exceptions.EncordException(
                f"An error has occurred during image group creation. {res.errors=}, {upload_job_id=}"
            )
        else:
            raise ValueError(f"{res.status=}, this should never happen, {upload_job_id=}")

    def create_dicom_series(
        self,
        file_paths: Union[typing.Collection[str], typing.Collection[Path], typing.Collection[Union[str, Path]]],
        title: Optional[str] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> Dict:
        """This function is documented in :meth:`encord.dataset.Dataset.create_dicom_series`."""
        signed_urls = upload_to_signed_url_list(
            file_paths=file_paths,
            config=self._config,
            querier=self._querier,
            orm_class=DicomSeries,
            cloud_upload_settings=cloud_upload_settings,
        )

        if not signed_urls:
            raise encord.exceptions.EncordException("DICOM files upload failed. The DICOM series was not created.")

        upload_job_id = self._querier.basic_setter(
            DatasetDataLongPolling,
            uid=self._querier.resource_id,
            payload=DatasetDataLongPollingParams(
                data_items=DataUploadItems(
                    dicom_series=[
                        DataUploadDicomSeries(
                            dicom_files=[
                                DataUploadDicomSeriesDicomFile(
                                    url=x["file_link"],
                                    title=x["title"],
                                )
                                for x in signed_urls
                            ],
                            title=title,
                        )
                    ],
                ),
                files=None,
                integration_id=None,
                ignore_errors=False,
                folder_uuid=folder_uuid,
                file_name=None,
            ),
        )["process_hash"]

        res = self.__add_data_to_dataset_get_result(upload_job_id)

        if res.status == LongPollingStatus.DONE:
            if len(res.data_hashes_with_titles) != 1:
                raise encord.exceptions.EncordException(
                    f"An error has occurred during the DICOM series creation. len({res.data_hashes_with_titles=}) != 1, {upload_job_id=}"
                )

            res_item = res.data_hashes_with_titles[0]

            return {
                "data_hash": res_item.data_hash,
                "title": res_item.title,
            }
        elif res.status == LongPollingStatus.ERROR:
            raise encord.exceptions.EncordException(
                f"An error has occurred during the DICOM series creation. {res.errors=}, {upload_job_id=}"
            )
        else:
            raise ValueError(f"{res.status=}, this should never happen, {upload_job_id=}")

    def upload_image(
        self,
        file_path: Union[Path, str],
        title: Optional[str] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> Image:
        """This function is documented in :meth:`encord.dataset.Dataset.upload_image`."""
        signed_urls = upload_to_signed_url_list(
            file_paths=[file_path],
            config=self._config,
            querier=self._querier,
            orm_class=Images,
            cloud_upload_settings=cloud_upload_settings,
        )

        if not signed_urls:
            raise encord.exceptions.EncordException("Image upload failed.")

        signed_url = signed_urls[0]

        upload_job_id = self._querier.basic_setter(
            DatasetDataLongPolling,
            uid=self._querier.resource_id,
            payload=DatasetDataLongPollingParams(
                data_items=DataUploadItems(
                    images=[
                        DataUploadImage(
                            object_url=signed_url["file_link"],
                            title=title or signed_url["title"],
                        )
                    ],
                ),
                files=None,
                integration_id=None,
                ignore_errors=False,
                folder_uuid=folder_uuid,
                file_name=None,
            ),
        )["process_hash"]

        res = self.__add_data_to_dataset_get_result(upload_job_id)

        if res.status == LongPollingStatus.DONE:
            if len(res.data_hashes_with_titles) != 1:
                raise encord.exceptions.EncordException(
                    f"An error has occurred during the image upload. len({res.data_hashes_with_titles=}) != 1, {upload_job_id=}"
                )

            res_item = res.data_hashes_with_titles[0]

            return Image(
                {
                    "data_hash": res_item.data_hash,
                    "title": res_item.title,
                    "file_link": signed_url["file_link"],
                }
            )
        elif res.status == LongPollingStatus.ERROR:
            raise encord.exceptions.EncordException(
                f"An error has occurred during the image upload. {res.errors=}, {upload_job_id=}"
            )
        else:
            raise ValueError(f"{res.status=}, this should never happen, {upload_job_id=}")

    def link_items(
        self,
        item_uuids: List[uuid.UUID],
        duplicates_behavior: DataLinkDuplicatesBehavior = DataLinkDuplicatesBehavior.SKIP,
    ) -> List[DataRow]:
        """Link storage items to the dataset, creating new data rows.

        Args:
            item_uuids: List of item UUIDs to link to the dataset
            duplicates_behavior: The behavior to follow when encountering duplicates. Defaults to `SKIP`. See also
                :class:`encord.orm.dataset.DataLinkDuplicatesBehavior`
        """
        data_row_dicts = self._querier.basic_setter(
            DatasetLinkItems,
            uid=self._querier.resource_id,
            payload={
                "item_uuids": [str(item_uuid) for item_uuid in item_uuids],
                "duplicates_behavior": duplicates_behavior.value,
            },
        )
        return DataRow.from_dict_list(data_row_dicts)

    def delete_data(self, data_hashes: Union[List[str], str]):
        """This function is documented in :meth:`encord.dataset.Dataset.delete_data`."""
        if isinstance(data_hashes, str):
            data_hashes = [data_hashes]
        self._querier.basic_delete(Video, uid=data_hashes)

    def add_private_data_to_dataset(
        self,
        integration_id: str,
        private_files: Union[str, typing.Dict, Path, typing.TextIO],
        ignore_errors: bool = False,
    ) -> AddPrivateDataResponse:
        """This function is documented in :meth:`encord.dataset.Dataset.add_private_data_to_dataset`."""
        upload_job_id = self.add_private_data_to_dataset_start(
            integration_id,
            private_files,
            ignore_errors,
        )

        res = self.add_private_data_to_dataset_get_result(upload_job_id)

        if res.status == LongPollingStatus.DONE:
            return AddPrivateDataResponse(dataset_data_list=res.data_hashes_with_titles)
        elif res.status == LongPollingStatus.ERROR:
            raise encord.exceptions.EncordException(
                f"add_private_data_to_dataset errors occurred {res.errors=}, {upload_job_id=}"
            )
        else:
            raise ValueError(f"{res.status=}, this should never happen, {upload_job_id=}")

    def add_private_data_to_dataset_start(
        self,
        integration_id: str,
        private_files: Union[str, typing.Dict, Path, typing.TextIO],
        ignore_errors: bool = False,
        folder_uuid: Optional[uuid.UUID] = None,
    ) -> str:
        """This function is documented in :meth:`encord.dataset.Dataset.add_private_data_to_dataset_start`."""
        if isinstance(private_files, dict):
            files = private_files
            file_name: Optional[str] = None
        elif isinstance(private_files, str) and os.path.exists(private_files):
            text_contents = Path(private_files).read_text(encoding="utf-8")
            files = json.loads(text_contents)
            file_name = Path(private_files).name
        elif isinstance(private_files, str) and (not os.path.exists(private_files)):
            text_contents = private_files
            files = json.loads(text_contents)
            file_name = None
        elif isinstance(private_files, Path):
            text_contents = private_files.read_text(encoding="utf-8")
            files = json.loads(text_contents)
            file_name = private_files.name
        elif isinstance(private_files, typing.TextIO):
            text_contents = private_files.read()
            files = json.loads(text_contents)
            file_name = None
        else:
            raise ValueError(f"Type [{type(private_files)}] of argument private_files is not supported")

        process_hash = self._querier.basic_setter(
            DatasetDataLongPolling,
            uid=self._querier.resource_id,
            payload=DatasetDataLongPollingParams(
                data_items=None,
                files=files,
                integration_id=UUID(integration_id),
                ignore_errors=ignore_errors,
                folder_uuid=folder_uuid,
                file_name=file_name,
            ),
        )["process_hash"]

        logger.info(f"add_private_data_to_dataset job started with upload_job_id={process_hash}.")
        logger.info("SDK process can be terminated, this will not affect successful job execution.")
        logger.info("You can follow the progress in the web app via notifications.")

        return process_hash

    def add_private_data_to_dataset_get_result(
        self,
        upload_job_id: str,
        timeout_seconds: int = 7 * 24 * 60 * 60,  # 7 days
    ) -> DatasetDataLongPolling:
        """This function is documented in :meth:`encord.dataset.Dataset.add_private_data_to_dataset_get_result`."""
        failed_requests_count = 0
        polling_start_timestamp = time.perf_counter()

        while True:
            try:
                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                res = self._querier.basic_getter(
                    DatasetDataLongPolling,
                    self._querier.resource_id,
                    payload={
                        "process_hash": upload_job_id,
                        "timeout_seconds": min(
                            polling_available_seconds,
                            LONG_POLLING_MAX_REQUEST_TIME_SECONDS,
                        ),
                    },
                )

                if res.status == LongPollingStatus.DONE:
                    logger.info(f"add_private_data_to_dataset job completed with upload_job_id={upload_job_id}.")

                polling_elapsed_seconds = ceil(time.perf_counter() - polling_start_timestamp)
                polling_available_seconds = max(0, timeout_seconds - polling_elapsed_seconds)

                if (polling_available_seconds == 0) or (
                    res.status
                    in [
                        LongPollingStatus.DONE,
                        LongPollingStatus.ERROR,
                        LongPollingStatus.CANCELLED,
                    ]
                ):
                    return res

                files_finished_count = res.units_done_count + res.units_error_count + res.units_cancelled_count
                files_total_count = (
                    res.units_pending_count + res.units_done_count + res.units_error_count + res.units_cancelled_count
                )

                if files_finished_count != files_total_count:
                    logger.info(f"Processed {files_finished_count}/{files_total_count} files")
                else:
                    logger.info("Processed all files, dataset data linking and task creation is performed, please wait")

                failed_requests_count = 0
            except (requests.exceptions.RequestException, encord.exceptions.RequestException):
                failed_requests_count += 1

                if failed_requests_count >= LONG_POLLING_RESPONSE_RETRY_N:
                    raise

                time.sleep(LONG_POLLING_SLEEP_ON_FAILURE_SECONDS)

    def update_data_item(self, data_hash: str, new_title: str) -> bool:
        """This function is documented in :meth:`encord.dataset.Dataset.update_data_item`."""
        payload = [{"video_hash": data_hash, "title": new_title}]

        response = self._querier.basic_setter(Video, self.get_dataset().dataset_hash, payload=payload)
        try:
            return response.get("success", False)
        except AttributeError:
            return False

    def re_encode_data(self, data_hashes: List[str]):
        """This function is documented in :meth:`encord.dataset.Dataset.re_encode_data`."""
        payload = {"data_hash": data_hashes}
        return self._querier.basic_put(ReEncodeVideoTask, uid=None, payload=payload)

    def re_encode_data_status(self, job_id: int):
        """This function is documented in :meth:`encord.dataset.Dataset.re_encode_data_status`."""
        return self._querier.basic_getter(ReEncodeVideoTask, uid=job_id)

    def run_ocr(self, image_group_id: str) -> List[ImageGroupOCR]:
        """This function is documented in :meth:`encord.dataset.Dataset.run_ocr`."""
        payload = {"image_group_data_hash": image_group_id}
        return self._querier.get_multiple(ImageGroupOCR, payload=payload)


class EncordClientProject(EncordClient):
    """DEPRECATED - prefer using the :class:`encord.project.Project` instead"""

    @property
    def project_hash(self) -> str:
        assert self._querier.resource_id, "Resource id can't be empty for created project client"
        return self._querier.resource_id  # type: ignore[attr-defined]

    def get_project(self, include_labels_metadata=True) -> OrmProject:
        """Retrieve project info (pointers to data, labels).

        Args:
            include_labels_metadata: if false, label row metadata information will not be returned.

        Returns:
            OrmProject: A project record instance.

        Raises:
            AuthorisationError: If the project API key is invalid.
            ResourceNotFoundError: If no project exists by the specified project EntityId.
            UnknownError: If an error occurs while retrieving the project.
        """
        return self._querier.basic_getter(
            OrmProject, payload={"include_labels_metadata": include_labels_metadata}, retryable=True
        )

    def get_project_v2(self) -> ProjectOrmV2:
        """This is an internal method, do not use it directly.
        Use :meth:`UserClient.get_project` instead.
        """
        return self._api_client.get(f"/projects/{self.project_hash}", params=None, result_type=ProjectOrmV2)

    def list_label_rows(
        self,
        edited_before: Optional[Union[str, datetime]] = None,
        edited_after: Optional[Union[str, datetime]] = None,
        label_statuses: Optional[List[AnnotationTaskStatus]] = None,
        shadow_data_state: Optional[ShadowDataState] = None,
        *,
        include_uninitialised_labels: bool = False,
        include_workflow_graph_node: bool = True,
        include_client_metadata: bool = False,
        include_images_data: bool = False,
        label_hashes: Optional[Union[List[str], List[UUID]]] = None,
        data_hashes: Optional[Union[List[str], List[UUID]]] = None,
        data_title_eq: Optional[str] = None,
        data_title_like: Optional[str] = None,
        workflow_graph_node_title_eq: Optional[str] = None,
        workflow_graph_node_title_like: Optional[str] = None,
        include_all_label_branches: bool = False,
        branch_name: Optional[str] = None,
    ) -> List[LabelRowMetadata]:
        """This function is documented in :meth:`encord.project.Project.list_label_rows`."""
        data_hashes = [str(data_hash) for data_hash in data_hashes] if data_hashes is not None else None
        label_hashes = [str(label_hash) for label_hash in label_hashes] if label_hashes is not None else None

        label_statuses_values = (
            [label_status.value for label_status in label_statuses] if label_statuses is not None else None
        )
        edited_before = optional_datetime_to_iso_str("edited_before", edited_before)
        edited_after = optional_datetime_to_iso_str("edited_after", edited_after)

        payload = {
            "edited_before": edited_before,
            "edited_after": edited_after,
            "label_statuses": label_statuses_values,
            "shadow_data_state": shadow_data_state.value if shadow_data_state else None,
            "include_uninitialised_labels": include_uninitialised_labels,
            "data_hashes": data_hashes,
            "label_hashes": label_hashes,
            "data_title_eq": data_title_eq,
            "data_title_like": data_title_like,
            "workflow_graph_node_title_eq": workflow_graph_node_title_eq,
            "workflow_graph_node_title_like": workflow_graph_node_title_like,
            "include_client_metadata": include_client_metadata,
            "include_images_data": include_images_data,
            "include_workflow_graph_node": include_workflow_graph_node,
            "include_all_label_branches": include_all_label_branches,
            "branch_name": branch_name,
        }
        return self._querier.get_multiple(LabelRowMetadata, payload=payload, retryable=True)

    def set_label_status(self, label_hash: str, label_status: LabelStatus) -> bool:
        """This function is documented in :meth:`encord.project.Project.set_label_status`."""
        payload = {
            "label_status": label_status.value,
        }
        return self._querier.basic_setter(LabelStatus, label_hash, payload)

    def add_users(self, user_emails: List[str], user_role: ProjectUserRole) -> List[ProjectUser]:
        """This function is documented in :meth:`encord.project.Project.add_users`."""
        payload = {"user_emails": user_emails, "user_role": user_role}
        users = self._querier.basic_setter(ProjectUsers, self._querier.resource_id, payload=payload)

        return [ProjectUser.from_dict(user) for user in users]

    def list_groups(self, project_hash: uuid.UUID) -> Page[ProjectGroup]:
        return self._api_client.get(f"projects/{project_hash}/groups", params=None, result_type=Page[ProjectGroup])

    def add_groups(self, project_hash: uuid.UUID, group_hash: List[uuid.UUID], user_role: ProjectUserRole) -> None:
        payload = AddProjectGroupsPayload(group_hash_list=group_hash, user_role=user_role)
        self._api_client.post(f"projects/{project_hash}/groups", params=None, payload=payload, result_type=None)

    def remove_groups(self, group_hash: List[uuid.UUID]) -> None:
        params = RemoveGroupsParams(group_hash_list=group_hash)
        self._api_client.delete(f"projects/{self.project_hash}/groups", params=params, result_type=None)

    def copy_project(
        self,
        copy_datasets: Union[bool, CopyDatasetOptions] = False,
        copy_collaborators=False,
        copy_models=False,
        *,
        copy_labels: Optional[CopyLabelsOptions] = None,
        new_title: Optional[str] = None,
        new_description: Optional[str] = None,
    ) -> str:
        """This function is documented in :meth:`encord.project.Project.copy_project`."""
        payload = CopyProjectPayload()
        payload.project_copy_metadata = CopyProjectPayload._ProjectCopyMetadata(new_title)
        payload.copy_labels_options = CopyProjectPayload._CopyLabelsOptions()

        if payload.project_copy_metadata and new_description:
            payload.project_copy_metadata.project_description = new_description

        if copy_labels:
            payload.copy_project_options.append(ProjectCopyOptions.LABELS)
            if isinstance(copy_labels, CopyLabelsOptions):
                payload.copy_labels_options.accepted_label_hashes = copy_labels.accepted_label_hashes
                payload.copy_labels_options.accepted_label_statuses = copy_labels.accepted_label_statuses

        if copy_datasets:
            if copy_datasets is True or copy_datasets.action == CopyDatasetAction.ATTACH:
                payload.copy_project_options.append(ProjectCopyOptions.DATASETS)
            elif copy_datasets.action == CopyDatasetAction.CLONE:
                payload.project_copy_metadata.dataset_title = copy_datasets.dataset_title
                payload.project_copy_metadata.dataset_description = copy_datasets.dataset_description
                payload.copy_labels_options.create_new_dataset = True
                payload.copy_labels_options.datasets_to_data_hashes_map = copy_datasets.datasets_to_data_hashes_map

        if copy_models:
            payload.copy_project_options.append(ProjectCopyOptions.MODELS)
        if copy_collaborators:
            payload.copy_project_options.append(ProjectCopyOptions.COLLABORATORS)

        return self._querier.basic_setter(ProjectCopy, self._querier.resource_id, payload=dataclasses.asdict(payload))

    def get_label_row(
        self,
        uid: str,
        get_signed_url: bool = True,
        *,
        include_object_feature_hashes: Optional[typing.Set[str]] = None,
        include_classification_feature_hashes: Optional[typing.Set[str]] = None,
        include_reviews: bool = False,
        include_export_history: bool = False,
    ) -> LabelRow:
        """This function is documented in :meth:`encord.project.Project.get_label_row`."""
        payload = {
            "get_signed_url": get_signed_url,
            "multi_request": False,
            "include_object_feature_hashes": optional_set_to_list(include_object_feature_hashes),
            "include_classification_feature_hashes": optional_set_to_list(include_classification_feature_hashes),
            "include_reviews": include_reviews,
            "include_export_history": include_export_history,
        }

        return self._querier.basic_getter(LabelRow, uid, payload=payload, retryable=True)

    def get_label_rows(
        self,
        uids: List[str],
        get_signed_url: bool = True,
        *,
        include_object_feature_hashes: Optional[typing.Set[str]] = None,
        include_classification_feature_hashes: Optional[typing.Set[str]] = None,
        include_reviews: bool = False,
        include_export_history: bool = False,
    ) -> List[LabelRow]:
        """This function is documented in :meth:`encord.project.Project.get_label_rows`."""
        payload = {
            "get_signed_url": get_signed_url,
            "multi_request": True,
            "include_object_feature_hashes": optional_set_to_list(include_object_feature_hashes),
            "include_classification_feature_hashes": optional_set_to_list(include_classification_feature_hashes),
            "include_reviews": include_reviews,
            "include_export_history": include_export_history,
        }

        return self._querier.get_multiple(LabelRow, uids, payload=payload, retryable=True)

    def save_label_row(self, uid, label, validate_before_saving: bool = False):
        """This function is documented in :meth:`encord.project.Project.save_label_row`."""
        label = LabelRow(label)
        if validate_before_saving:
            label["validate_before_saving"] = True
        return self._querier.basic_setter(LabelRow, uid, payload=label, retryable=True)

    def save_label_rows(self, uids: List[str], payload: List[LabelRow], validate_before_saving: bool = False):
        """This function is meant for internal use, please consider using :class:`encord.objects.LabelRowV2` class instead

        Saves multiple label rows. See :meth:`.save_label_row`

        Args:
            uids: list of label hashes
            payload: list of LabelRow objects
            validate_before_saving: ????

        Returns:
            None
        """
        multirequest_payload = {
            "multi_request": True,
            "labels": payload,
            "validate_before_saving": validate_before_saving,
        }
        return self._querier.basic_setter(LabelRow, uid=uids, payload=multirequest_payload, retryable=True)

    def create_label_row(self, uid, *, get_signed_url=False) -> LabelRow:
        """This function is documented in :meth:`encord.project.Project.create_label_row`."""
        return self._querier.basic_put(LabelRow, uid=uid, payload={"get_signed_url": get_signed_url})

    def create_label_rows(
        self, uids: List[str], *, get_signed_url: bool = False, branch_name: Optional[str] = None
    ) -> List[LabelRow]:
        """This function is meant for internal use, please consider using :class:`encord.objects.LabelRowV2` class instead

        Create multiple label rows. See :meth:`.create_label_row`

        Args:
            uids: list of data uids where label_status is NOT_LABELLED.
            get_signed_url: bool whether to fetch the signed url for the internal label row
            branch_name: Optional[str] which branch name against which to create the label row
        Returns:
            List[LabelRow]: A list of created label rows
        """
        return self._querier.put_multiple(
            LabelRow,
            uid=uids,
            payload={"multi_request": True, "get_signed_url": get_signed_url, "branch_name": branch_name},
        )

    def submit_label_row_for_review(self, uid):
        """This function is documented in :meth:`encord.project.Project.submit_label_row_for_review`."""
        return self._querier.basic_put(Review, uid=uid, payload=None)

    def add_datasets(self, dataset_hashes: List[str]) -> bool:
        """This function is documented in :meth:`encord.project.Project.add_datasets`."""
        payload = {"dataset_hashes": dataset_hashes}
        return self._querier.basic_setter(ProjectDataset, uid=None, payload=payload)

    def remove_datasets(self, dataset_hashes: List[str]) -> bool:
        """This function is documented in :meth:`encord.project.Project.remove_datasets`."""
        return self._querier.basic_delete(ProjectDataset, uid=dataset_hashes)

    def list_project_datasets(self, project_hash: UUID) -> Iterable[ProjectDataset]:
        return self._api_client.get(
            f"projects/{project_hash}/datasets", params=None, result_type=Page[ProjectDataset]
        ).results

    @deprecated("0.1.102", alternative="encord.ontology.Ontology class")
    def get_project_ontology(self) -> LegacyOntology:
        project = self.get_project()
        ontology = project["editor_ontology"]
        return LegacyOntology.from_dict(ontology)

    @deprecated("0.1.102", alternative="encord.ontology.Ontology class")
    def add_object(self, name: str, shape: ObjectShape) -> bool:
        """This function is documented in :meth:`encord.project.Project.add_object`."""
        if not name:
            raise ValueError("Ontology object name is empty")

        ontology = self.get_project_ontology()
        ontology.add_object(name, shape)
        return self.__set_project_ontology(ontology)

    @deprecated("0.1.102", alternative="encord.ontology.Ontology class")
    def add_classification(
        self,
        name: str,
        classification_type: ClassificationType,
        required: bool,
        options: Optional[Iterable[str]] = None,
    ):
        """This function is documented in :meth:`encord.project.Project.add_classification`."""
        if not name:
            raise ValueError("Ontology classification name is empty")

        ontology = self.get_project_ontology()
        ontology.add_classification(name, classification_type, required, options)
        return self.__set_project_ontology(ontology)

    def object_interpolation(
        self,
        key_frames,
        objects_to_interpolate,
    ):
        """This function is documented in :meth:`encord.project.Project.object_interpolation`."""
        if len(key_frames) == 0 or len(objects_to_interpolate) == 0:
            raise encord.exceptions.EncordException(
                message="To run object interpolation, you must pass key frames and objects to interpolate.."
            )

        interpolation_params = ObjectInterpolationParams(
            {
                "key_frames": key_frames,
                "objects_to_interpolate": objects_to_interpolate,
            }
        )

        algo = LabelingAlgorithm(
            {
                "algorithm_name": INTERPOLATION,
                "algorithm_parameters": interpolation_params,
            }
        )

        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)

    def fitted_bounding_boxes(
        self,
        frames: dict,
        video: dict,
    ):
        """This function is documented in :meth:`encord.project.Project.fitted_bounding_boxes`."""
        if len(frames) == 0 or len(video) == 0:
            raise encord.exceptions.EncordException(
                message="To run fitting, you must pass frames and video to run bounding box fitting on.."
            )

        fitting_params = BoundingBoxFittingParams(
            {
                "labels": frames,
                "video": video,
            }
        )

        algo = LabelingAlgorithm(
            {
                "algorithm_name": FITTED_BOUNDING_BOX,
                "algorithm_parameters": fitting_params,
            }
        )

        return self._querier.basic_setter(LabelingAlgorithm, str(uuid.uuid4()), payload=algo)

    def get_data(self, data_hash: str, get_signed_url: bool = False) -> Tuple[Optional[Video], Optional[List[Image]]]:
        """This function is documented in :meth:`encord.project.Project.get_data`."""
        uid = {"data_hash": data_hash, "get_signed_url": get_signed_url}

        dataset_data: DatasetData = self._querier.basic_getter(DatasetData, uid=uid)

        video: Union[Video, None] = None
        if dataset_data["video"] is not None:
            video = Video(dataset_data["video"])

        images: Union[List[Image], None] = None
        if dataset_data["images"] is not None:
            images = []

            for image in dataset_data["images"]:
                images.append(Image(image))

        return video, images

    def get_label_logs(
        self,
        user_hash: Optional[str] = None,
        data_hash: Optional[str] = None,
        from_unix_seconds: Optional[int] = None,
        to_unix_seconds: Optional[int] = None,
        after: Optional[datetime] = None,
        before: Optional[datetime] = None,
        user_email: Optional[str] = None,
    ) -> List[LabelLog]:
        """This function is documented in :meth:`encord.project.Project.get_label_logs`."""
        if after is not None:
            if from_unix_seconds is not None:
                raise ValueError("Only one of 'from_unix_seconds' and 'after' parameters should be specified")

            from_unix_seconds = int(after.timestamp())

        if before is not None:
            if to_unix_seconds is not None:
                raise ValueError("Only one of 'to_unix_seconds' and 'before' parameters should be specified")

            to_unix_seconds = int(before.timestamp())

        payload = LabelLogParams(
            user_hash=user_hash,
            user_email=user_email,
            data_hash=data_hash,
            end_timestamp=to_unix_seconds,
            start_timestamp=from_unix_seconds,
        )

        return self._querier.get_multiple(LabelLog, payload=payload.to_dict(by_alias=False))

    def __set_project_ontology(self, ontology: LegacyOntology) -> bool:
        """Save updated project ontology
        Args:
            ontology: the updated project ontology

        Returns:
            bool
        """
        payload = {"editor": ontology.to_dict()}
        return self._querier.basic_setter(OrmProject, uid=None, payload=payload)

    def workflow_reopen(self, label_hashes: List[str]) -> None:
        """This function is documented in :meth:`encord.objects.LabelRowV2.workflow_reopen`."""
        self._querier.basic_setter(
            LabelWorkflowGraphNode,
            label_hashes,
            payload=LabelWorkflowGraphNodePayload(action=WorkflowAction.REOPEN),
        )

    def workflow_complete(self, label_hashes: List[str]) -> None:
        """This function is documented in :meth:`encord.objects.LabelRowV2.workflow_complete`."""
        self._querier.basic_setter(
            LabelWorkflowGraphNode,
            label_hashes,
            payload=LabelWorkflowGraphNodePayload(action=WorkflowAction.COMPLETE),
        )

    def workflow_set_priority(self, priorities: List[Tuple[str, float]]) -> None:
        self._api_client.post(
            f"projects/{self.project_hash}/priorities",
            params=None,
            payload=TaskPriorityParams(priorities=priorities),
            result_type=None,
        )

    def get_collaborator_timers(self, params: CollaboratorTimerParams) -> Iterable[CollaboratorTimer]:
        yield from self._api_client.get_paged_iterator(
            "analytics/collaborators/timers", params=params, result_type=CollaboratorTimer
        )

    def get_time_spent(self, params: TimeSpentParams) -> Iterable[TimeSpent]:
        yield from self._api_client.get_paged_iterator("analytics/time-spent", params=params, result_type=TimeSpent)

    def get_label_validation_errors(self, label_hash: str) -> List[str]:
        errors = self._api_client.get(
            f"projects/{self.project_hash}/labels/{label_hash}/validation-state",
            params=None,
            result_type=LabelValidationState,
        )

        if errors.is_valid:
            return []

        return errors.errors or []

    def active_import(self, project_mode: ActiveProjectMode, video_sampling_rate: Optional[float] = None) -> None:
        self._api_client.post(
            f"active/{self.project_hash}/import",
            params=None,
            payload=ActiveProjectImportPayload(project_mode=project_mode, video_sampling_rate=video_sampling_rate),
            result_type=None,
        )
        logger.info("Import initiated in Active, please check the app to see progress")

    def active_sync(self) -> None:
        self._api_client.post(
            f"active/{self.project_hash}/sync",
            params=None,
            payload=None,
            result_type=None,
        )
        logger.info("Sync initiated in Active, please check the app to see progress")


CordClientProject = EncordClientProject



from __future__ import annotations

import datetime
from collections import OrderedDict
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union
from uuid import UUID

from encord.exceptions import WrongProjectTypeError
from encord.orm import base_orm
from encord.orm.analytics import CamelStrEnum
from encord.orm.base_dto import BaseDTO
from encord.orm.workflow import Workflow
from encord.utilities.project_user import ProjectUserRole


class Project(base_orm.BaseORM):
    """DEPRECATED - prefer using the `encord.project.Project` class instead.

    A project defines a label ontology and is a collection of datasets and label rows.

    ORM:

    * title,
    * description,
    * editor_ontology,
    * ontology_hash,
    * datasets::

        [
           {
                dataset_hash (uid),
                title,
                description,
                dataset_type (internal vs. AWS/GCP/Azure),
           },
           ...
        ],
    * label_rows::

        [
            {
                label_hash (uid),
                data_hash (uid),
                dataset_hash (uid),
                dataset_title,
                data_title,
                data_type,
                label_status
            },
            ...
        ]
    * annotation_task_status

    """

    DB_FIELDS = OrderedDict(
        [
            ("project_hash", str),
            ("title", str),
            ("description", str),
            ("created_at", datetime.datetime),
            ("last_edited_at", datetime.datetime),
            ("editor_ontology", (dict, str)),
            ("datasets", (list, str)),
            ("label_rows", (list, str)),
            ("ontology_hash", str),
            ("source_projects", list),
            ("workflow_manager_uuid", str),
        ]
    )

    NON_UPDATABLE_FIELDS = {"editor_ontology", "datasets", "label_rows"}

    def get_labels_list(self) -> List[Optional[str]]:
        """Returns a list of all optional label row IDs (label_hash uid) in a project. If no `label_hash` is found,
        a `None` value is appended. This can be useful for working with fetching additional label row data via
        :meth:`encord.project.Project.get_label_rows` for example.

        .. code::

                project = client_instance.get_project(<project_hash>)
                project_orm = project.get_project()

                labels_list = project_orm.get_labels_list()
                created_labels_list = []
                for label in labels_list:
                    if label is not None:  # None values will fail the operation
                        created_labels_list.append(label)

                label_rows = project.get_label_rows(created_labels_list, get_signed_url=False)
        """
        labels = self.label_rows() or []
        return [label.get("label_hash") for label in labels]

    @property
    def project_hash(self) -> str:
        return self["project_hash"]

    @property
    def title(self) -> str:
        return self["title"]

    @property
    def description(self) -> str:
        return self["description"]

    @property
    def editor_ontology(self):
        return self["editor_ontology"]

    @property
    def datasets(self):
        return self["datasets"]

    @property
    def label_rows(self):
        return self["label_rows"]

    @property
    def ontology_hash(self) -> str:
        return self["ontology_hash"]

    @property
    def source_projects(self):
        return self["source_projects"]

    @property
    def created_at(self) -> datetime.datetime:
        return self["created_at"]

    @property
    def last_edited_at(self) -> datetime.datetime:
        return self["last_edited_at"]

    @property
    def workflow_manager_uuid(self) -> UUID:
        """Accessing this property will raise a `WrongProjectTypeError` if the project is not a workflow project."""
        try:
            return self["workflow_manager_uuid"]
        except KeyError as e:
            raise WrongProjectTypeError(
                "This project is not a workflow project, workflow_manager_uuid is not available."
            ) from e


class ProjectCopy:
    pass


class ProjectUsers:
    pass


class ProjectDataset(BaseDTO):
    dataset_hash: UUID
    title: str
    description: str

    @classmethod
    def from_list(cls, dataset_list: List[Dict[str, Any]]) -> List[ProjectDataset]:
        return [cls.from_dict(i) for i in dataset_list]


class ProjectType(str, Enum):
    WORKFLOW = "workflow"
    MANUAL_QA = "manual_qa"


class ProjectCopyOptions(str, Enum):
    COLLABORATORS = "collaborators"
    DATASETS = "datasets"
    MODELS = "models"
    LABELS = "labels"


class ReviewApprovalState(str, Enum):
    APPROVED = "APPROVED"
    PENDING = "PENDING"
    REJECTED = "REJECTED"
    DELETED = "DELETED"
    NOT_SELECTED_FOR_REVIEW = "NOT_SELECTED_FOR_REVIEW"


class CopyDatasetAction(str, Enum):
    ATTACH = "ATTACH"
    """ Attach the datasets associated with the original project to the copy project. """
    CLONE = "CLONE"
    """ Clone the data units from the associated datasets into a new dataset an attach it to the copy project. """


@dataclass
class CopyDatasetOptions:
    """Options for copying the datasets associated with a project."""

    action: CopyDatasetAction = CopyDatasetAction.ATTACH
    """
    One of `CopyDatasetAction.ATTACH` or `CopyDatasetAction.CLONE`. (defaults to ATTACH)
    """
    dataset_title: Optional[str] = None
    dataset_description: Optional[str] = None
    datasets_to_data_hashes_map: Dict[str, List[str]] = field(default_factory=dict)
    """ A dictionary of `{ <dataset_hash>: List[<data_unit_hash>]}`.
    When provided with a CLONE action this will filter the copied data units.
    When combined with `CopyLabelsOptions`, only labels from specific data units will be copied.
    """


@dataclass
class CopyLabelsOptions:
    """Options for copying the labels associated with a project."""

    accepted_label_hashes: Optional[List[str]] = None
    """ A list of label hashes that will be copied to the new project  """
    accepted_label_statuses: Optional[List[ReviewApprovalState]] = None
    """ A list of label statuses to filter the labels copied to the new project, defined in `ReviewApprovalState`"""


@dataclass
class CopyProjectPayload:
    """WARN: do not use, this is only for internal purpose."""

    @dataclass
    class _CopyLabelsOptions:
        datasets_to_data_hashes_map: Dict[str, List[str]] = field(default_factory=dict)
        accepted_label_hashes: Optional[List[str]] = None
        accepted_label_statuses: Optional[List[ReviewApprovalState]] = None
        create_new_dataset: Optional[bool] = None

    @dataclass
    class _ProjectCopyMetadata:
        project_title: Optional[str] = None
        project_description: Optional[str] = None
        dataset_title: Optional[str] = None
        dataset_description: Optional[str] = None

    copy_project_options: List[ProjectCopyOptions] = field(default_factory=list)
    copy_labels_options: Optional[_CopyLabelsOptions] = None
    project_copy_metadata: Optional[_ProjectCopyMetadata] = None


class ProjectWorkflowType(Enum):
    MANUAL_QA = "manual"
    BENCHMARK_QA = "benchmark"


class ManualReviewWorkflowSettings:
    """Sets the project QA workflow to "manual reviews". There are currently no available settings for this workflow"""

    pass


@dataclass
class BenchmarkQaWorkflowSettings:
    """Sets the project QA workflow to "Automatic", with benchmark data being presented to all the annotators"""

    source_projects: List[str] = field(default_factory=list)
    """
    For Benchmark QA projects, a list of project ids (project_hash-es)
        that contain the benchmark source data
    """


ProjectWorkflowSettings = Union[ManualReviewWorkflowSettings, BenchmarkQaWorkflowSettings]
"""
A variant type that allows you to select the type of quality assurance workflow for the project,
and configure it.

Currently one of:
    :class:`ManualReviewWorkflowSettings`: a workflow with optional manual reviews
    :class:`BenchmarkQaWorkflowSettings`: annotators are presented with "benchmark" or "honeypot" data
"""


class ReviewMode(str, Enum):
    """UNLABELLED:
        The labels are added to the images. However, the one person must still go over
            all of the labels before submitting them for review.
    LABELLED:
        The labels are added to the images and are marked as labelled. A reviewer will
            still need to approve those.
    REVIEWED:
        The labels are added to the images and considered reviewed. No more action is
            required from the labeler or reviewer.
    """

    UNLABELLED = "unlabelled"
    LABELLED = "labelled"
    REVIEWED = "reviewed"


class CvatExportType(str, Enum):
    PROJECT = "project"
    TASK = "task"


class TaskPriorityParams(BaseDTO):
    priorities: List[Tuple[str, float]]


class ProjectDTO(BaseDTO):
    project_hash: UUID
    project_type: ProjectType
    title: str
    description: str
    created_at: datetime.datetime
    last_edited_at: datetime.datetime
    ontology_hash: str
    editor_ontology: Dict[str, Any]
    user_role: Optional[ProjectUserRole] = None
    source_projects: Optional[List[str]] = None
    workflow_manager_uuid: Optional[UUID] = None
    workflow: Optional[Workflow] = None


class CvatReviewMode(CamelStrEnum):
    """UNLABELLED:
        The labels are added to the images. However, the one person must still go over
            all the labels before submitting them for review.
    LABELLED:
        The labels are added to the images and are marked as labelled. A reviewer will
            still need to approve those.
    REVIEWED:
        The labels are added to the images and considered reviewed. No more action is
            required from the labeler or reviewer.
    """

    UNLABELLED = auto()
    LABELLED = auto()
    REVIEWED = auto()


class CvatImportDataItem(BaseDTO):
    data_path: str
    data_link: str
    title: str


class CvatImportStartPayload(BaseDTO):
    annotations_base64: str
    dataset_uuid: UUID
    review_mode: CvatReviewMode
    data: List[CvatImportDataItem]
    transform_bounding_boxes_to_polygons: bool


class CvatImportGetResultParams(BaseDTO):
    timeout_seconds: int


class CvatImportGetResultLongPollingStatus(str, Enum):
    DONE = "DONE"
    ERROR = "ERROR"
    PENDING = "PENDING"


class CvatImportGetResultResponse(BaseDTO):
    """Response model for CVAT import operation status.

    Attributes:
        status (CvatImportGetResultLongPollingStatus): Import status ("DONE", "ERROR", "PENDING")
        project_uuid (Optional[UUID]): Created project ID. Only present when status="DONE"
        issues (Optional[Dict]): Import issues. Present for "DONE" (warnings) or "ERROR" (failures)
    """

    status: CvatImportGetResultLongPollingStatus
    project_uuid: Optional[UUID] = None
    issues: Optional[Dict] = None


class ProjectFilterParams(BaseDTO):
    """Filter parameters for the /v2/public/projects endpoint"""

    title_eq: Optional[str] = None
    title_like: Optional[str] = None
    desc_eq: Optional[str] = None
    desc_like: Optional[str] = None
    created_before: Optional[Union[str, datetime.datetime]] = None
    created_after: Optional[Union[str, datetime.datetime]] = None
    edited_before: Optional[Union[str, datetime.datetime]] = None
    edited_after: Optional[Union[str, datetime.datetime]] = None
    include_org_access: bool = False
